# Copyright (C) 2012 - 2021  Christian Ledermann
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA

"""Frequently used constants and configuration options"""
import logging
import warnings
from types import ModuleType

__all__ = [
    "ATOMNS",
    "DEFAULT_NAME_SPACES",
    "FORCE3D",
    "GXNS",
    "KMLNS",
    "register_namespaces",
    "set_default_namespaces",
    "set_etree_implementation",
]

try:  # pragma: no cover
    from lxml import etree

except ImportError:  # pragma: no cover
    warnings.warn("Package `lxml` missing. Pretty print will be disabled")
    import xml.etree.ElementTree as etree  # type: ignore[no-redef] # noqa: N813


logger = logging.getLogger(__name__)


def set_etree_implementation(implementation: ModuleType) -> None:
    """Set the etree implementation to use."""
    global etree
    etree = implementation


KMLNS = "{http://www.opengis.net/kml/2.2}"  # noqa: FS003
ATOMNS = "{http://www.w3.org/2005/Atom}"  # noqa: FS003
GXNS = "{http://www.google.com/kml/ext/2.2}"  # noqa: FS003

DEFAULT_NAME_SPACES = {
    "kml": KMLNS[1:-1],
    "atom": ATOMNS[1:-1],
    "gx": GXNS[1:-1],
}


def register_namespaces(**namespaces: str) -> None:
    """Register namespaces for use in etree.ElementTree.parse()."""
    try:
        for prefix, uri in namespaces.items():
            etree.register_namespace(prefix, uri)
    except AttributeError:  # pragma: no cover
        logger.warning("Namespaces were not registered.")


def set_default_namespaces() -> None:
    """Register the default namespaces for use in etree.ElementTree.parse()."""
    register_namespaces(**DEFAULT_NAME_SPACES)


set_default_namespaces()

FORCE3D = False



# Copyright (C) 2021  Christian Ledermann
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA

"""Test the base classes."""
from typing import cast

import pytest

from fastkml import base
from fastkml import config
from fastkml import types
from tests.base import Lxml
from tests.base import StdLibrary


class TestStdLibrary(StdLibrary):
    """Test the base object with the standard library."""

    def test_to_string(self) -> None:
        obj = base._BaseObject(id="id-0", target_id="target-id-0")
        obj.__name__ = "test"

        assert (
            obj.to_string()
            == '<kml:test xmlns:kml="http://www.opengis.net/kml/2.2" id="id-0" targetId="target-id-0" />'
        )

    def test_from_string(self) -> None:
        be = base._BaseObject()
        be.__name__ = "test"

        be.from_string(
            xml_string='<kml:test xmlns:kml="http://www.opengis.net/kml/2.2" id="id-0" targetId="target-id-0" />'
        )

        assert be.id == "id-0"
        assert be.target_id == "target-id-0"

    def test_base_etree_element_raises(self) -> None:
        be = base._BaseObject()

        with pytest.raises(NotImplementedError):
            be.etree_element()

    def test_base_etree_element_raises_subclass(self) -> None:
        class Test(base._BaseObject):
            pass

        with pytest.raises(NotImplementedError):
            Test().etree_element()

    def test_base_from_element_raises(self) -> None:
        be = base._BaseObject()
        element = cast(types.Element, config.etree.Element(config.KMLNS + "Base"))  # type: ignore[attr-defined]

        with pytest.raises(TypeError):
            be.from_element(element=element)

    def test_base_from_string_raises(self) -> None:
        be = base._BaseObject()

        with pytest.raises(TypeError):
            be.from_string(
                xml_string='<kml:test xmlns:kml="http://www.opengis.net/kml/2.2" id="id-0" />'
            )


class TestLxml(Lxml, TestStdLibrary):
    """Test the base object with lxml."""

    def test_to_string(self) -> None:
        obj = base._BaseObject(id="id-0")
        obj.__name__ = "test"

        assert (
            obj.to_string()
            == '<kml:test xmlns:kml="http://www.opengis.net/kml/2.2" id="id-0"/>\n'
        )

    def test_from_string(self) -> None:
        be = base._BaseObject()
        be.__name__ = "test"

        be.from_string(
            xml_string='<kml:test xmlns:kml="http://www.opengis.net/kml/2.2" id="id-0" targetId="target-id-0"/>\n'
        )

        assert be.id == "id-0"
        assert be.target_id == "target-id-0"



import logging
from typing import Optional
from typing import Union

import fastkml.config as config
import fastkml.gx as gx
from fastkml.base import _BaseObject
from fastkml.mixins import TimeMixin
from fastkml.times import TimeSpan
from fastkml.times import TimeStamp
from fastkml.types import Element

logger = logging.getLogger(__name__)


class _AbstractView(TimeMixin, _BaseObject):
    """
    This is an abstract element and cannot be used directly in a KML file.
    This element is extended by the <Camera> and <LookAt> elements.
    """

    _longitude: Optional[float] = None
    # Longitude of the virtual camera (eye point). Angular distance in degrees,
    # relative to the Prime Meridian. Values west of the Meridian range from
    # −180 to 0 degrees. Values east of the Meridian range from 0 to 180 degrees.

    _latitude: Optional[float] = None
    # Latitude of the virtual camera. Degrees north or south of the Equator
    # (0 degrees). Values range from −90 degrees to 90 degrees.

    _altitude: Optional[float] = None
    # Distance of the camera from the earth's surface, in meters. Interpreted
    # according to the Camera's <altitudeMode> or <gx:altitudeMode>.

    _heading: Optional[float] = None
    # Direction (azimuth) of the camera, in degrees. Default=0 (true North).
    # (See diagram.) Values range from 0 to 360 degrees.

    _tilt: Optional[float] = None
    # Rotation, in degrees, of the camera around the X axis. A value of 0
    # indicates that the view is aimed straight down toward the earth (the
    # most common case). A value for 90 for <tilt> indicates that the view
    # is aimed toward the horizon. Values greater than 90 indicate that the
    # view is pointed up into the sky. Values for <tilt> are clamped at +180
    # degrees.

    _altitude_mode: str = "relativeToGround"
    # Specifies how the <altitude> specified for the Camera is interpreted.
    # Possible values are as follows:
    #   relativeToGround -
    #       (default) Interprets the <altitude> as a value in meters above the
    #       ground. If the point is over water, the <altitude> will be
    #       interpreted as a value in meters above sea level. See
    #       <gx:altitudeMode> below to specify points relative to the sea floor.
    #   clampToGround -
    #       For a camera, this setting also places the camera relativeToGround,
    #       since putting the camera exactly at terrain height would mean that
    #       the eye would intersect the terrain (and the view would be blocked).
    #   absolute -
    #       Interprets the <altitude> as a value in meters above sea level.

    def __init__(
        self,
        ns: Optional[str] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        longitude: Optional[float] = None,
        latitude: Optional[float] = None,
        altitude: Optional[float] = None,
        heading: Optional[float] = None,
        tilt: Optional[float] = None,
        altitude_mode: str = "relativeToGround",
        time_primitive: Union[TimeSpan, TimeStamp, None] = None,
    ) -> None:
        super().__init__(ns=ns, id=id, target_id=target_id)
        self._longitude = longitude
        self._latitude = latitude
        self._altitude = altitude
        self._heading = heading
        self._tilt = tilt
        self._altitude_mode = altitude_mode
        if isinstance(time_primitive, TimeSpan):
            self._timespan = time_primitive
        elif isinstance(time_primitive, TimeStamp):
            self._timestamp = time_primitive

    @property
    def longitude(self) -> Optional[float]:
        return self._longitude

    @longitude.setter
    def longitude(self, value) -> None:
        if isinstance(value, (str, int, float)) and (-180 <= float(value) <= 180):
            self._longitude = float(value)
        elif value is None:
            self._longitude = None
        else:
            raise ValueError

    @property
    def latitude(self) -> Optional[float]:
        return self._latitude

    @latitude.setter
    def latitude(self, value) -> None:
        if isinstance(value, (str, int, float)) and (-90 <= float(value) <= 90):
            self._latitude = float(value)
        elif value is None:
            self._latitude = None
        else:
            raise ValueError

    @property
    def altitude(self) -> Optional[float]:
        return self._altitude

    @altitude.setter
    def altitude(self, value) -> None:
        if isinstance(value, (str, int, float)):
            self._altitude = float(value)
        elif value is None:
            self._altitude = None
        else:
            raise ValueError

    @property
    def heading(self) -> Optional[float]:
        return self._heading

    @heading.setter
    def heading(self, value) -> None:
        if isinstance(value, (str, int, float)) and (-180 <= float(value) <= 180):
            self._heading = float(value)
        elif value is None:
            self._heading = None
        else:
            raise ValueError

    @property
    def tilt(self) -> Optional[float]:
        return self._tilt

    @tilt.setter
    def tilt(self, value) -> None:
        if isinstance(value, (str, int, float)) and (0 <= float(value) <= 180):
            self._tilt = float(value)
        elif value is None:
            self._tilt = None
        else:
            raise ValueError

    @property
    def altitude_mode(self) -> str:
        return self._altitude_mode

    @altitude_mode.setter
    def altitude_mode(self, mode: str) -> None:
        if mode in {"relativeToGround", "clampToGround", "absolute"}:
            self._altitude_mode = mode
        else:
            self._altitude_mode = "relativeToGround"
            # raise ValueError(
            #     "altitude_mode must be one of " "relativeToGround,
            #     clampToGround, absolute")

    def from_element(self, element: Element):
        super().from_element(element)
        longitude = element.find(f"{self.ns}longitude")
        if longitude is not None:
            self.longitude = longitude.text
        latitude = element.find(f"{self.ns}latitude")
        if latitude is not None:
            self.latitude = latitude.text
        altitude = element.find(f"{self.ns}altitude")
        if altitude is not None:
            self.altitude = altitude.text
        heading = element.find(f"{self.ns}heading")
        if heading is not None:
            self.heading = heading.text
        tilt = element.find(f"{self.ns}tilt")
        if tilt is not None:
            self.tilt = tilt.text
        altitude_mode = element.find(f"{self.ns}altitudeMode")
        if altitude_mode is None:
            altitude_mode = element.find(f"{gx.NS}altitudeMode")
        self.altitude_mode = altitude_mode.text
        timespan = element.find(f"{self.ns}TimeSpan")
        if timespan is not None:
            s = TimeSpan(self.ns)
            s.from_element(timespan)
            self._timespan = s
        timestamp = element.find(f"{self.ns}TimeStamp")
        if timestamp is not None:
            s = TimeStamp(self.ns)
            s.from_element(timestamp)
            self._timestamp = s

    def etree_element(self) -> Element:
        element = super().etree_element()
        if self.longitude:
            longitude = config.etree.SubElement(element, f"{self.ns}longitude")
            longitude.text = str(self.longitude)
        if self.latitude:
            latitude = config.etree.SubElement(element, f"{self.ns}latitude")
            latitude.text = str(self.latitude)
        if self.altitude:
            altitude = config.etree.SubElement(element, f"{self.ns}altitude")
            altitude.text = str(self.altitude)
        if self.heading:
            heading = config.etree.SubElement(element, f"{self.ns}heading")
            heading.text = str(self.heading)
        if self.tilt:
            tilt = config.etree.SubElement(element, f"{self.ns}tilt")
            tilt.text = str(self.tilt)
        if self.altitude_mode in ("clampedToGround", "relativeToGround", "absolute"):
            altitude_mode = config.etree.SubElement(element, f"{self.ns}altitudeMode")
        elif self.altitude_mode in ("clampedToSeaFloor", "relativeToSeaFloor"):
            altitude_mode = config.etree.SubElement(element, f"{gx.NS}altitudeMode")
        altitude_mode.text = self.altitude_mode
        if (self._timespan is not None) and (self._timestamp is not None):
            raise ValueError("Either Timestamp or Timespan can be defined, not both")
        if self._timespan is not None:
            element.append(self._timespan.etree_element())
        elif self._timestamp is not None:
            element.append(self._timestamp.etree_element())
        return element

    # TODO: <gx:ViewerOptions>
    # TODO: <gx:horizFov>


class Camera(_AbstractView):
    """
    Defines the virtual camera that views the scene. This element defines
    the position of the camera relative to the Earth's surface as well
    as the viewing direction of the camera. The camera position is defined
    by <longitude>, <latitude>, <altitude>, and either <altitudeMode> or
    <gx:altitudeMode>. The viewing direction of the camera is defined by
    <heading>, <tilt>, and <roll>. <Camera> can be a child element of any
    Feature or of <NetworkLinkControl>. A parent element cannot contain both a
    <Camera> and a <LookAt> at the same time.

    <Camera> provides full six-degrees-of-freedom control over the view,
    so you can position the Camera in space and then rotate it around the
    X, Y, and Z axes. Most importantly, you can tilt the camera view so that
    you're looking above the horizon into the sky.

    <Camera> can also contain a TimePrimitive (<gx:TimeSpan> or <gx:TimeStamp>).
    Time values in Camera affect historical imagery, sunlight, and the display of
    time-stamped features. For more information, read Time with AbstractViews in
    the Time and Animation chapter of the Developer's Guide.
    """

    __name__ = "Camera"

    _roll: Optional[float] = None
    # Rotation, in degrees, of the camera around the Z axis. Values range from
    # −180 to +180 degrees.

    def __init__(
        self,
        ns: Optional[str] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        longitude: Optional[float] = None,
        latitude: Optional[float] = None,
        altitude: Optional[float] = None,
        heading: Optional[float] = None,
        tilt: Optional[float] = None,
        roll: Optional[float] = None,
        altitude_mode: str = "relativeToGround",
        time_primitive: Union[TimeSpan, TimeStamp, None] = None,
    ) -> None:
        super().__init__(
            ns=ns,
            id=id,
            target_id=target_id,
            longitude=longitude,
            latitude=latitude,
            altitude=altitude,
            heading=heading,
            tilt=tilt,
            altitude_mode=altitude_mode,
            time_primitive=time_primitive,
        )
        self._roll = roll

    def from_element(self, element: Element) -> None:
        super().from_element(element)
        roll = element.find(f"{self.ns}roll")
        if roll is not None:
            self.roll = roll.text

    def etree_element(self) -> Element:
        element = super().etree_element()
        if self.roll:
            roll = config.etree.SubElement(element, f"{self.ns}roll")
            roll.text = str(self.roll)
        return element

    @property
    def roll(self) -> Optional[float]:
        return self._roll

    @roll.setter
    def roll(self, value) -> None:
        if isinstance(value, (str, int, float)) and (-180 <= float(value) <= 180):
            self._roll = float(value)
        elif value is None:
            self._roll = None
        else:
            raise ValueError


class LookAt(_AbstractView):

    __name__ = "LookAt"

    _range: Optional[float] = None
    # Distance in meters from the point specified by <longitude>, <latitude>,
    # and <altitude> to the LookAt position. (See diagram below.)

    def __init__(
        self,
        ns: Optional[str] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        longitude: Optional[float] = None,
        latitude: Optional[float] = None,
        altitude: Optional[float] = None,
        heading: Optional[float] = None,
        tilt: Optional[float] = None,
        range: Optional[float] = None,
        altitude_mode: str = "relativeToGround",
        time_primitive: Union[TimeSpan, TimeStamp, None] = None,
    ) -> None:
        super().__init__(
            ns=ns,
            id=id,
            target_id=target_id,
            longitude=longitude,
            latitude=latitude,
            altitude=altitude,
            heading=heading,
            tilt=tilt,
            altitude_mode=altitude_mode,
            time_primitive=time_primitive,
        )
        self._range = range

    @property
    def range(self) -> Optional[float]:
        return self._range

    @range.setter
    def range(self, value) -> None:
        if isinstance(value, (str, int, float)):
            self._range = float(value)
        elif value is None:
            self._range = None
        else:
            raise ValueError

    def from_element(self, element: Element) -> None:
        super().from_element(element)
        range_var = element.find(f"{self.ns}range")
        if range_var is not None:
            self.range = range_var.text

    def etree_element(self) -> Element:
        element = super().etree_element()
        if self.range:
            range_var = config.etree.SubElement(element, f"{self.ns}range")
            range_var.text = str(self._range)
        return element


__all__ = [
    "Camera",
    "LookAt",
]



# Copyright (C) 2012-2022  Christian Ledermann
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA

"""
KML is an open standard officially named the OpenGIS KML Encoding Standard
(OGC KML). It is maintained by the Open Geospatial Consortium, Inc. (OGC).
The complete specification for OGC KML can be found at
http://www.opengeospatial.org/standards/kml/.

The complete XML schema for KML is located at
http://schemas.opengis.net/kml/.

"""
import logging
import urllib.parse as urlparse
from typing import Iterator
from typing import List
from typing import Optional
from typing import Union

import fastkml.atom as atom
import fastkml.config as config
import fastkml.gx as gx
from fastkml.base import _BaseObject
from fastkml.data import ExtendedData
from fastkml.data import Schema
from fastkml.geometry import Geometry
from fastkml.mixins import TimeMixin
from fastkml.styles import Style
from fastkml.styles import StyleMap
from fastkml.styles import StyleUrl
from fastkml.styles import _StyleSelector
from fastkml.times import TimeSpan
from fastkml.times import TimeStamp
from fastkml.types import Element
from fastkml.views import Camera
from fastkml.views import LookAt

logger = logging.getLogger(__name__)


class _Feature(TimeMixin, _BaseObject):
    """
    abstract element; do not create
    subclasses are:
        * Container (Document, Folder)
        * Placemark
        * Overlay
    Not Implemented Yet:
        * NetworkLink
    """

    name = None
    # User-defined text displayed in the 3D viewer as the label for the
    # object (for example, for a Placemark, Folder, or NetworkLink).

    visibility = 1
    # Boolean value. Specifies whether the feature is drawn in the 3D
    # viewer when it is initially loaded. In order for a feature to be
    # visible, the <visibility> tag of all its ancestors must also be
    # set to 1.

    isopen = 0
    # Boolean value. Specifies whether a Document or Folder appears
    # closed or open when first loaded into the Places panel.
    # 0=collapsed (the default), 1=expanded.

    _atom_author = None
    # KML 2.2 supports new elements for including data about the author
    # and related website in your KML file. This information is displayed
    # in geo search results, both in Earth browsers such as Google Earth,
    # and in other applications such as Google Maps.

    _atom_link = None
    # Specifies the URL of the website containing this KML or KMZ file.

    _address = None
    # A string value representing an unstructured address written as a
    # standard street, city, state address, and/or as a postal code.
    # You can use the <address> tag to specify the location of a point
    # instead of using latitude and longitude coordinates.

    _phone_number = None
    # A string value representing a telephone number.
    # This element is used by Google Maps Mobile only.

    _snippet = None  # XXX
    # _snippet is either a tuple of a string Snippet.text and an integer
    # Snippet.maxLines or a string
    #
    # A short description of the feature. In Google Earth, this
    # description is displayed in the Places panel under the name of the
    # feature. If a Snippet is not supplied, the first two lines of
    # the <description> are used. In Google Earth, if a Placemark
    # contains both a description and a Snippet, the <Snippet> appears
    # beneath the Placemark in the Places panel, and the <description>
    # appears in the Placemark's description balloon. This tag does not
    # support HTML markup. <Snippet> has a maxLines attribute, an integer
    # that specifies the maximum number of lines to display.

    description = None
    # User-supplied content that appears in the description balloon.

    _style_url = None
    # URL of a <Style> or <StyleMap> defined in a Document.
    # If the style is in the same file, use a # reference.
    # If the style is defined in an external file, use a full URL
    # along with # referencing.

    _styles = None
    # One or more Styles and StyleMaps can be defined to customize the
    # appearance of any element derived from Feature or of the Geometry
    # in a Placemark.
    # A style defined within a Feature is called an "inline style" and
    # applies only to the Feature that contains it. A style defined as
    # the child of a <Document> is called a "shared style." A shared
    # style must have an id defined for it. This id is referenced by one
    # or more Features within the <Document>. In cases where a style
    # element is defined both in a shared style and in an inline style
    # for a Feature—that is, a Folder, GroundOverlay, NetworkLink,
    # Placemark, or ScreenOverlay—the value for the Feature's inline
    # style takes precedence over the value for the shared style.

    _timespan = None
    # Associates this Feature with a period of time.
    _timestamp = None
    # Associates this Feature with a point in time.

    _camera = None

    _look_at = None

    # TODO Region = None
    # Features and geometry associated with a Region are drawn only when
    # the Region is active.

    # TODO ExtendedData = None
    # Allows you to add custom data to a KML file. This data can be
    # (1) data that references an external XML schema,
    # (2) untyped data/value pairs, or
    # (3) typed data.
    # A given KML Feature can contain a combination of these types of
    # custom data.
    #
    # (2) is already implemented, see UntypedExtendedData
    #
    # <Metadata> (deprecated in KML 2.2; use <ExtendedData> instead)
    extended_data = None

    def __init__(
        self,
        ns: Optional[str] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        styles: Optional[List[Style]] = None,
        style_url: Optional[str] = None,
        extended_data: None = None,
    ) -> None:
        super().__init__(ns=ns, id=id, target_id=target_id)
        self.name = name
        self.description = description
        self.style_url = style_url
        self._styles = []
        if styles:
            for style in styles:
                self.append_style(style)
        self.extended_data = extended_data

    @property
    def style_url(self) -> Optional[str]:
        """Returns the url only, not a full StyleUrl object.
        if you need the full StyleUrl object use _style_url"""
        if isinstance(self._style_url, StyleUrl):
            return self._style_url.url

    @style_url.setter
    def style_url(self, styleurl: Union[str, StyleUrl, None]) -> None:
        """you may pass a StyleUrl Object, a string or None"""
        if isinstance(styleurl, StyleUrl):
            self._style_url = styleurl
        elif isinstance(styleurl, str):
            s = StyleUrl(self.ns, url=styleurl)
            self._style_url = s
        elif styleurl is None:
            self._style_url = None
        else:
            raise ValueError

    @property
    def camera(self):
        return self._camera

    @camera.setter
    def camera(self, camera):
        if isinstance(camera, Camera):
            self._camera = camera

    @property
    def look_at(self):
        return self._look_at

    @look_at.setter
    def look_at(self, look_at):
        if isinstance(look_at, LookAt):
            self._look_at = look_at

    @property
    def link(self):
        return self._atom_link.href

    @link.setter
    def link(self, url):
        if isinstance(url, str):
            self._atom_link = atom.Link(href=url)
        elif isinstance(url, atom.Link):
            self._atom_link = url
        elif url is None:
            self._atom_link = None
        else:
            raise TypeError

    @property
    def author(self):
        if self._atom_author:
            return self._atom_author.name

    @author.setter
    def author(self, name):
        if isinstance(name, atom.Author):
            self._atom_author = name
        elif isinstance(name, str):
            if self._atom_author is None:
                self._atom_author = atom.Author(name=name)
            else:
                self._atom_author.name = name
        elif name is None:
            self._atom_author = None
        else:
            raise TypeError

    def append_style(self, style: Union[Style, StyleMap]) -> None:
        """append a style to the feature"""
        if isinstance(style, _StyleSelector):
            self._styles.append(style)
        else:
            raise TypeError

    def styles(self) -> Iterator[Union[Style, StyleMap]]:
        """iterate over the styles of this feature"""
        for style in self._styles:
            if isinstance(style, _StyleSelector):
                yield style
            else:
                raise TypeError

    @property
    def snippet(self):
        if not self._snippet:
            return
        if isinstance(self._snippet, dict):
            text = self._snippet.get("text")
            if text:
                assert isinstance(text, str)
                max_lines = self._snippet.get("maxLines", None)
                if max_lines is None:
                    return {"text": text}
                elif int(max_lines) > 0:
                    # if maxLines <=0 ignore it
                    return {"text": text, "maxLines": max_lines}
        elif isinstance(self._snippet, str):
            return self._snippet
        else:
            raise ValueError(
                "Snippet must be dict of "
                "{'text':t, 'maxLines':i} or string"  # noqa: FS003
            )

    @snippet.setter
    def snippet(self, snip=None):
        self._snippet = {}
        if isinstance(snip, dict):
            self._snippet["text"] = snip.get("text")
            max_lines = snip.get("maxLines")
            if max_lines is not None:
                self._snippet["maxLines"] = int(snip["maxLines"])
        elif isinstance(snip, str):
            self._snippet["text"] = snip
        elif snip is None:
            self._snippet = None
        else:
            raise ValueError(
                "Snippet must be dict of "
                "{'text':t, 'maxLines':i} or string"  # noqa: FS003
            )

    @property
    def address(self):
        if self._address:
            return self._address

    @address.setter
    def address(self, address):
        if isinstance(address, str):
            self._address = address
        elif address is None:
            self._address = None
        else:
            raise ValueError

    @property
    def phone_number(self):
        if self._phone_number:
            return self._phone_number

    @phone_number.setter
    def phone_number(self, phone_number):
        if isinstance(phone_number, str):
            self._phone_number = phone_number
        elif phone_number is None:
            self._phone_number = None
        else:
            raise ValueError

    def etree_element(self) -> Element:
        element = super().etree_element()
        if self.name:
            name = config.etree.SubElement(element, f"{self.ns}name")
            name.text = self.name
        if self.description:
            description = config.etree.SubElement(element, f"{self.ns}description")
            description.text = self.description
        if (self.camera is not None) and (self.look_at is not None):
            raise ValueError("Either Camera or LookAt can be defined, not both")
        if self.camera is not None:
            element.append(self._camera.etree_element())
        elif self.look_at is not None:
            element.append(self._look_at.etree_element())
        visibility = config.etree.SubElement(element, f"{self.ns}visibility")
        visibility.text = str(self.visibility)
        if self.isopen:
            isopen = config.etree.SubElement(element, f"{self.ns}open")
            isopen.text = str(self.isopen)
        if self._style_url is not None:
            element.append(self._style_url.etree_element())
        for style in self.styles():
            element.append(style.etree_element())
        if self.snippet:
            snippet = config.etree.SubElement(element, f"{self.ns}Snippet")
            if isinstance(self.snippet, str):
                snippet.text = self.snippet
            else:
                assert isinstance(self.snippet["text"], str)
                snippet.text = self.snippet["text"]
                if self.snippet.get("maxLines"):
                    snippet.set("maxLines", str(self.snippet["maxLines"]))
        if (self._timespan is not None) and (self._timestamp is not None):
            raise ValueError("Either Timestamp or Timespan can be defined, not both")
        elif self._timespan is not None:
            element.append(self._timespan.etree_element())
        elif self._timestamp is not None:
            element.append(self._timestamp.etree_element())
        if self._atom_link is not None:
            element.append(self._atom_link.etree_element())
        if self._atom_author is not None:
            element.append(self._atom_author.etree_element())
        if self.extended_data is not None:
            element.append(self.extended_data.etree_element())
        if self._address is not None:
            address = config.etree.SubElement(element, f"{self.ns}address")
            address.text = self._address
        if self._phone_number is not None:
            phone_number = config.etree.SubElement(element, f"{self.ns}phoneNumber")
            phone_number.text = self._phone_number
        return element

    def from_element(self, element: Element) -> None:
        super().from_element(element)
        name = element.find(f"{self.ns}name")
        if name is not None:
            self.name = name.text
        description = element.find(f"{self.ns}description")
        if description is not None:
            self.description = description.text
        visibility = element.find(f"{self.ns}visibility")
        if visibility is not None:
            self.visibility = 1 if visibility.text in ["1", "true"] else 0
        isopen = element.find(f"{self.ns}open")
        if isopen is not None:
            self.isopen = 1 if isopen.text in ["1", "true"] else 0
        styles = element.findall(f"{self.ns}Style")
        for style in styles:
            s = Style(self.ns)
            s.from_element(style)
            self.append_style(s)
        styles = element.findall(f"{self.ns}StyleMap")
        for style in styles:
            s = StyleMap(self.ns)
            s.from_element(style)
            self.append_style(s)
        style_url = element.find(f"{self.ns}styleUrl")
        if style_url is not None:
            s = StyleUrl(self.ns)
            s.from_element(style_url)
            self._style_url = s
        snippet = element.find(f"{self.ns}Snippet")
        if snippet is not None:
            _snippet = {"text": snippet.text}
            if snippet.get("maxLines"):
                _snippet["maxLines"] = int(snippet.get("maxLines"))
            self.snippet = _snippet
        timespan = element.find(f"{self.ns}TimeSpan")
        if timespan is not None:
            s = TimeSpan(self.ns)
            s.from_element(timespan)
            self._timespan = s
        timestamp = element.find(f"{self.ns}TimeStamp")
        if timestamp is not None:
            s = TimeStamp(self.ns)
            s.from_element(timestamp)
            self._timestamp = s
        atom_link = element.find(f"{atom.NS}link")
        if atom_link is not None:
            s = atom.Link()
            s.from_element(atom_link)
            self._atom_link = s
        atom_author = element.find(f"{atom.NS}author")
        if atom_author is not None:
            s = atom.Author()
            s.from_element(atom_author)
            self._atom_author = s
        extended_data = element.find(f"{self.ns}ExtendedData")
        if extended_data is not None:
            x = ExtendedData(self.ns)
            x.from_element(extended_data)
            self.extended_data = x
            # else:
            #    logger.warn(
            #        'arbitrary or typed extended data is not yet supported'
            #    )
        address = element.find(f"{self.ns}address")
        if address is not None:
            self.address = address.text
        phone_number = element.find(f"{self.ns}phoneNumber")
        if phone_number is not None:
            self.phone_number = phone_number.text
        camera = element.find(f"{self.ns}Camera")
        if camera is not None:
            s = Camera(self.ns)
            s.from_element(camera)
            self.camera = s


class Icon(_BaseObject):
    """
    Represents an <Icon> element used in Overlays.

    Defines an image associated with an Icon style or overlay.
    The required <href> child element defines the location
    of the image to be used as the overlay or as the icon for the placemark.
    This location can either be on a local file system or a remote web server.
    The <gx:x>, <gx:y>, <gx:w>, and <gx:h> elements are used to select one
    icon from an image that contains multiple icons
    (often referred to as an icon palette).
    """

    __name__ = "Icon"

    _href = None
    _refresh_mode: str = None
    _refresh_interval = None
    _view_refresh_mode = None
    _view_refresh_time = None
    _view_bound_scale = None
    _view_format = None
    _http_query = None

    def __init__(
        self,
        ns: Optional[str] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        href: Optional[str] = None,
        refresh_mode: Optional[str] = None,
        refresh_interval: Optional[float] = None,
        view_refresh_mode: Optional[str] = None,
        view_refresh_time: Optional[float] = None,
        view_bound_scale: Optional[float] = None,
        view_format: Optional[str] = None,
        http_query: Optional[str] = None,
    ) -> None:
        """Initialize the KML Icon Object."""
        super().__init__(ns=ns, id=id, target_id=target_id)
        self._href = href
        self._refresh_mode = refresh_mode
        self._refresh_interval = refresh_interval
        self._view_refresh_mode = view_refresh_mode
        self._view_refresh_time = view_refresh_time
        self._view_bound_scale = view_bound_scale
        self._view_format = view_format
        self._http_query = http_query

    @property
    def href(self) -> Optional[str]:
        """An HTTP address or a local file specification used to load an icon."""
        return self._href

    @href.setter
    def href(self, href) -> None:
        if isinstance(href, str):
            self._href = href
        elif href is None:
            self._href = None
        else:
            raise ValueError

    @property
    def refresh_mode(self) -> Optional[str]:
        """
        Specifies a time-based refresh mode.

        Can be one of the following:
          - onChange - refresh when the file is loaded and whenever the Link
            parameters change (the default).
          - onInterval - refresh every n seconds (specified in <refreshInterval>).
          - onExpire - refresh the file when the expiration time is reached.
        """
        return self._refresh_mode

    @refresh_mode.setter
    def refresh_mode(self, refresh_mode) -> None:
        if isinstance(refresh_mode, str):
            self._refresh_mode = refresh_mode
        elif refresh_mode is None:
            self._refresh_mode = None
        else:
            raise ValueError

    @property
    def refresh_interval(self) -> Optional[float]:
        """Indicates to refresh the file every n seconds."""
        return self._refresh_interval

    @refresh_interval.setter
    def refresh_interval(self, refresh_interval: Optional[float]) -> None:
        if isinstance(refresh_interval, float):
            self._refresh_interval = refresh_interval
        elif refresh_interval is None:
            self._refresh_interval = None
        else:
            raise ValueError

    @property
    def view_refresh_mode(self):
        """
        Specifies how the link is refreshed when the "camera" changes.

        Can be one of the following:
         - never (default) - Ignore changes in the view.
           Also ignore <viewFormat> parameters, if any.
         - onStop - Refresh the file n seconds after movement stops,
           where n is specified in <viewRefreshTime>.
         - onRequest - Refresh the file only when the user explicitly requests it.
           (For example, in Google Earth, the user right-clicks and selects Refresh
            in the Context menu.)
          - onRegion - Refresh the file when the Region becomes active.
        """
        return self._view_refresh_mode

    @view_refresh_mode.setter
    def view_refresh_mode(self, view_refresh_mode):
        if isinstance(view_refresh_mode, str):
            self._view_refresh_mode = view_refresh_mode
        elif view_refresh_mode is None:
            self._view_refresh_mode = None
        else:
            raise ValueError

    @property
    def view_refresh_time(self):
        """
        After camera movement stops, specifies the number of seconds to
        wait before refreshing the view.
        """
        return self._view_refresh_time

    @view_refresh_time.setter
    def view_refresh_time(self, view_refresh_time: Optional[float]):
        if isinstance(view_refresh_time, float):
            self._view_refresh_time = view_refresh_time
        elif view_refresh_time is None:
            self._view_refresh_time = None
        else:
            raise ValueError

    @property
    def view_bound_scale(self) -> Optional[float]:
        """
        Scales the BBOX parameters before sending them to the server.

        A value less than 1 specifies to use less than the full view (screen).
        A value greater than 1 specifies to fetch an area that extends beyond
        the edges of the current view.
        """
        return self._view_bound_scale

    @view_bound_scale.setter
    def view_bound_scale(self, view_bound_scale: Optional[float]) -> None:
        if isinstance(view_bound_scale, float):
            self._view_bound_scale = view_bound_scale
        elif view_bound_scale is None:
            self._view_bound_scale = None
        else:
            raise ValueError

    @property
    def view_format(self):
        """
        Specifies the format of the query string that is appended to the
        Link's <href> before the file is fetched.
        (If the <href> specifies a local file, this element is ignored.)

        This information matches the Web Map Service (WMS) bounding box specification.
        If you specify an empty <viewFormat> tag, no information is appended to the
        query string.
        You can also specify a custom set of viewing parameters to add to the query
        string. If you supply a format string, it is used instead of the BBOX
        information. If you also want the BBOX information, you need to add those
        parameters along with the custom parameters.
        You can use any of the following parameters in your format string:
         - [lookatLon], [lookatLat] - longitude and latitude of the point that
           <LookAt> is viewing
         - [lookatRange], [lookatTilt], [lookatHeading] - values used by the
           <LookAt> element (see descriptions of <range>, <tilt>, and <heading>
           in <LookAt>)
         - [lookatTerrainLon], [lookatTerrainLat], [lookatTerrainAlt] - point on the
           terrain in degrees/meters that <LookAt> is viewing
         - [cameraLon], [cameraLat], [cameraAlt] - degrees/meters of the eyepoint for
           the camera
         - [horizFov], [vertFov] - horizontal, vertical field of view for the camera
         - [horizPixels], [vertPixels] - size in pixels of the 3D viewer
         - [terrainEnabled] - indicates whether the 3D viewer is showing terrain (1)
           or not (0)
        """
        return self._view_format

    @view_format.setter
    def view_format(self, view_format):
        if isinstance(view_format, str):
            self._view_format = view_format
        elif view_format is None:
            self._view_format = None
        else:
            raise ValueError

    @property
    def http_query(self):
        """
        Appends information to the query string, based on the parameters specified.

        The following parameters are supported:
         - [clientVersion]
         - [kmlVersion]
         - [clientName]
         - [language]
        """
        return self._http_query

    @http_query.setter
    def http_query(self, http_query):
        if isinstance(http_query, str):
            self._http_query = http_query
        elif http_query is None:
            self._http_query = None
        else:
            raise ValueError

    def etree_element(self) -> Element:
        element = super().etree_element()

        if self._href:
            href = config.etree.SubElement(element, f"{self.ns}href")
            href.text = self._href
        if self._refresh_mode:
            refresh_mode = config.etree.SubElement(element, f"{self.ns}refreshMode")
            refresh_mode.text = self._refresh_mode
        if self._refresh_interval:
            refresh_interval = config.etree.SubElement(
                element, f"{self.ns}refreshInterval"
            )
            refresh_interval.text = str(self._refresh_interval)
        if self._view_refresh_mode:
            view_refresh_mode = config.etree.SubElement(
                element, f"{self.ns}viewRefreshMode"
            )
            view_refresh_mode.text = self._view_refresh_mode
        if self._view_refresh_time:
            view_refresh_time = config.etree.SubElement(
                element, f"{self.ns}viewRefreshTime"
            )
            view_refresh_time.text = str(self._view_refresh_time)
        if self._view_bound_scale:
            view_bound_scale = config.etree.SubElement(
                element, f"{self.ns}viewBoundScale"
            )
            view_bound_scale.text = str(self._view_bound_scale)
        if self._view_format:
            view_format = config.etree.SubElement(element, f"{self.ns}viewFormat")
            view_format.text = self._view_format
        if self._http_query:
            http_query = config.etree.SubElement(element, f"{self.ns}httpQuery")
            http_query.text = self._http_query

        return element

    def from_element(self, element: Element) -> None:
        super().from_element(element)

        href = element.find(f"{self.ns}href")
        if href is not None:
            self.href = href.text

        refresh_mode = element.find(f"{self.ns}refreshMode")
        if refresh_mode is not None:
            self.refresh_mode = refresh_mode.text

        refresh_interval = element.find(f"{self.ns}refreshInterval")
        if refresh_interval is not None:
            try:
                self.refresh_interval = float(refresh_interval.text)
            except ValueError:
                self.refresh_interval = None

        view_refresh_mode = element.find(f"{self.ns}viewRefreshMode")
        if view_refresh_mode is not None:
            self.view_refresh_mode = view_refresh_mode.text

        view_refresh_time = element.find(f"{self.ns}viewRefreshTime")
        if view_refresh_time is not None:
            try:
                self.view_refresh_time = float(view_refresh_time.text)
            except ValueError:
                self.view_refresh_time = None

        view_bound_scale = element.find(f"{self.ns}viewBoundScale")
        if view_bound_scale is not None:
            try:
                self.view_bound_scale = float(view_bound_scale.text)
            except ValueError:
                self.view_bound_scale = None

        view_format = element.find(f"{self.ns}viewFormat")
        if view_format is not None:
            self.view_format = view_format.text

        http_query = element.find(f"{self.ns}httpQuery")
        if http_query is not None:
            self.http_query = http_query.text


class _Container(_Feature):
    """
    abstract element; do not create
    A Container element holds one or more Features and allows the
    creation of nested hierarchies.
    subclasses are:
    Document,
    Folder
    """

    _features = []

    def __init__(
        self,
        ns: Optional[str] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        styles: Optional[List[Style]] = None,
        style_url: Optional[str] = None,
        features: None = None,
    ) -> None:
        super().__init__(
            ns=ns,
            id=id,
            target_id=target_id,
            name=name,
            description=description,
            styles=styles,
            style_url=style_url,
        )
        self._features = features or []

    def features(self) -> Iterator[_Feature]:
        """iterate over features"""
        for feature in self._features:
            if isinstance(feature, (Folder, Placemark, Document, _Overlay)):
                yield feature
            else:
                raise TypeError(
                    "Features must be instances of "
                    "(Folder, Placemark, Document, Overlay)"
                )

    def etree_element(self) -> Element:
        element = super().etree_element()
        for feature in self.features():
            element.append(feature.etree_element())
        return element

    def append(self, kmlobj: _Feature) -> None:
        """append a feature"""
        if isinstance(kmlobj, (Folder, Placemark, Document, _Overlay)):
            self._features.append(kmlobj)
        else:
            raise TypeError(
                "Features must be instances of "
                "(Folder, Placemark, Document, Overlay)"
            )
        assert kmlobj != self


class _Overlay(_Feature):
    """
    abstract element; do not create

    Base type for image overlays drawn on the planet surface or on the screen

    A Container element holds one or more Features and allows the creation of
    nested hierarchies.
    """

    _color = None
    # Color values expressed in hexadecimal notation, including opacity (alpha)
    # values. The order of expression is alphaOverlay, blue, green, red
    # (AABBGGRR). The range of values for any one color is 0 to 255 (00 to ff).
    # For opacity, 00 is fully transparent and ff is fully opaque.

    _draw_order = None
    # Defines the stacking order for the images in overlapping overlays.
    # Overlays with higher <drawOrder> values are drawn on top of those with
    # lower <drawOrder> values.

    _icon = None
    # Defines the image associated with the overlay. Contains an <href> html
    # tag which defines the location of the image to be used as the overlay.
    # The location can be either on a local file system or on a webserver. If
    # this element is omitted or contains no <href>, a rectangle is drawn using
    # the color and size defined by the ground or screen overlay.

    def __init__(
        self,
        ns: Optional[str] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        styles: None = None,
        style_url: Optional[str] = None,
        icon: Optional[Icon] = None,
    ) -> None:
        super().__init__(
            ns=ns,
            id=id,
            target_id=target_id,
            name=name,
            description=description,
            styles=styles,
            style_url=style_url,
        )
        self._icon = icon

    @property
    def color(self):
        return self._color

    @color.setter
    def color(self, color):
        if isinstance(color, str):
            self._color = color
        elif color is None:
            self._color = None
        else:
            raise ValueError

    @property
    def draw_order(self):
        return self._draw_order

    @draw_order.setter
    def draw_order(self, value):
        if isinstance(value, (str, int, float)):
            self._draw_order = str(value)
        elif value is None:
            self._draw_order = None
        else:
            raise ValueError

    @property
    def icon(self):
        return self._icon

    @icon.setter
    def icon(self, value):
        if isinstance(value, Icon):
            self._icon = value
        elif value is None:
            self._icon = None
        else:
            raise ValueError

    def etree_element(self) -> Element:
        element = super().etree_element()
        if self._color:
            color = config.etree.SubElement(element, f"{self.ns}color")
            color.text = self._color
        if self._draw_order:
            draw_order = config.etree.SubElement(element, f"{self.ns}drawOrder")
            draw_order.text = self._draw_order
        if self._icon:
            element.append(self._icon.etree_element())
        return element

    def from_element(self, element: Element) -> None:
        super().from_element(element)
        color = element.find(f"{self.ns}color")
        if color is not None:
            self.color = color.text
        draw_order = element.find(f"{self.ns}drawOrder")
        if draw_order is not None:
            self.draw_order = draw_order.text
        icon = element.find(f"{self.ns}Icon")
        if icon is not None:
            s = Icon(self.ns)
            s.from_element(icon)
            self._icon = s


class PhotoOverlay(_Overlay):
    """
    The <PhotoOverlay> element allows you to geographically locate a photograph
    on the Earth and to specify viewing parameters for this PhotoOverlay.
    The PhotoOverlay can be a simple 2D rectangle, a partial or full cylinder,
    or a sphere (for spherical panoramas). The overlay is placed at the
    specified location and oriented toward the viewpoint.

    Because <PhotoOverlay> is derived from <Feature>, it can contain one of
    the two elements derived from <AbstractView>—either <Camera> or <LookAt>.
    The Camera (or LookAt) specifies a viewpoint and a viewing direction (also
    referred to as a view vector). The PhotoOverlay is positioned in relation
    to the viewpoint. Specifically, the plane of a 2D rectangular image is
    orthogonal (at right angles to) the view vector. The normal of this
    plane—that is, its front, which is the part
    with the photo—is oriented toward the viewpoint.

    The URL for the PhotoOverlay image is specified in the <Icon> tag,
    which is inherited from <Overlay>. The <Icon> tag must contain an <href>
    element that specifies the image file to use for the PhotoOverlay.
    In the case of a very large image, the <href> is a special URL that
    indexes into a pyramid of images of varying resolutions (see ImagePyramid).
    """

    __name__ = "PhotoOverlay"

    _rotation = None
    # Adjusts how the photo is placed inside the field of view. This element is
    # useful if your photo has been rotated and deviates slightly from a desired
    # horizontal view.

    # - ViewVolume -
    # Defines how much of the current scene is visible. Specifying the field of
    # view is analogous to specifying the lens opening in a physical camera.
    # A small field of view, like a telephoto lens, focuses on a small part of
    # the scene. A large field of view, like a wide-angle lens, focuses on a
    # large part of the scene.

    _left_fow = None
    # Angle, in degrees, between the camera's viewing direction and the left side
    # of the view volume.

    _right_fov = None
    # Angle, in degrees, between the camera's viewing direction and the right side
    # of the view volume.

    _bottom_fov = None
    # Angle, in degrees, between the camera's viewing direction and the bottom side
    # of the view volume.

    _top_fov = None
    # Angle, in degrees, between the camera's viewing direction and the top side
    # of the view volume.

    _near = None
    # Measurement in meters along the viewing direction from the camera viewpoint
    # to the PhotoOverlay shape.

    _tile_size = "256"
    # Size of the tiles, in pixels. Tiles must be square, and <tileSize> must
    # be a power of 2. A tile size of 256 (the default) or 512 is recommended.
    # The original image is divided into tiles of this size, at varying resolutions.

    _max_width = None
    # Width in pixels of the original image.

    _max_height = None
    # Height in pixels of the original image.

    _grid_origin = None
    # Specifies where to begin numbering the tiles in each layer of the pyramid.
    # A value of lowerLeft specifies that row 1, column 1 of each layer is in
    # the bottom left corner of the grid.

    _point = None
    # The <Point> element acts as a <Point> inside a <Placemark> element.
    # It draws an icon to mark the position of the PhotoOverlay.
    # The icon drawn is specified by the <styleUrl> and <StyleSelector> fields,
    # just as it is for <Placemark>.

    _shape = "rectangle"
    # The PhotoOverlay is projected onto the <shape>.
    # The <shape> can be one of the following:
    #   rectangle (default) -
    #       for an ordinary photo
    #   cylinder -
    #       for panoramas, which can be either partial or full cylinders
    #   sphere -
    #       for spherical panoramas

    @property
    def rotation(self):
        return self._rotation

    @rotation.setter
    def rotation(self, value):
        if isinstance(value, (str, int, float)):
            self._rotation = str(value)
        elif value is None:
            self._rotation = None
        else:
            raise ValueError

    @property
    def left_fov(self):
        return self._left_fow

    @left_fov.setter
    def left_fov(self, value):
        if isinstance(value, (str, int, float)):
            self._left_fow = str(value)
        elif value is None:
            self._left_fow = None
        else:
            raise ValueError

    @property
    def right_fov(self):
        return self._right_fov

    @right_fov.setter
    def right_fov(self, value):
        if isinstance(value, (str, int, float)):
            self._right_fov = str(value)
        elif value is None:
            self._right_fov = None
        else:
            raise ValueError

    @property
    def bottom_fov(self):
        return self._bottom_fov

    @bottom_fov.setter
    def bottom_fov(self, value):
        if isinstance(value, (str, int, float)):
            self._bottom_fov = str(value)
        elif value is None:
            self._bottom_fov = None
        else:
            raise ValueError

    @property
    def top_fov(self):
        return self._top_fov

    @top_fov.setter
    def top_fov(self, value):
        if isinstance(value, (str, int, float)):
            self._top_fov = str(value)
        elif value is None:
            self._top_fov = None
        else:
            raise ValueError

    @property
    def near(self):
        return self._near

    @near.setter
    def near(self, value):
        if isinstance(value, (str, int, float)):
            self._near = str(value)
        elif value is None:
            self._near = None
        else:
            raise ValueError

    @property
    def tile_size(self):
        return self._tile_size

    @tile_size.setter
    def tile_size(self, value):
        if isinstance(value, (str, int, float)):
            self._tile_size = str(value)
        elif value is None:
            self._tile_size = None
        else:
            raise ValueError

    @property
    def max_width(self):
        return self._max_width

    @max_width.setter
    def max_width(self, value):
        if isinstance(value, (str, int, float)):
            self._max_width = str(value)
        elif value is None:
            self._max_width = None
        else:
            raise ValueError

    @property
    def max_height(self):
        return self._max_height

    @max_height.setter
    def max_height(self, value):
        if isinstance(value, (str, int, float)):
            self._max_height = str(value)
        elif value is None:
            self._max_height = None
        else:
            raise ValueError

    @property
    def grid_origin(self):
        return self._grid_origin

    @grid_origin.setter
    def grid_origin(self, value):
        if value in ("lowerLeft", "upperLeft"):
            self._grid_origin = str(value)
        elif value is None:
            self._grid_origin = None
        else:
            raise ValueError

    @property
    def point(self):
        return self._point

    @point.setter
    def point(self, value):
        if isinstance(value, (str, tuple)):
            self._point = str(value)
        else:
            raise ValueError

    @property
    def shape(self):
        return self._shape

    @shape.setter
    def shape(self, value):
        if value in ("rectangle", "cylinder", "sphere"):
            self._shape = str(value)
        elif value is None:
            self._shape = None
        else:
            raise ValueError("Shape must be one of " "rectangle, cylinder, sphere")

    def view_volume(self, left_fov, right_fov, bottom_fov, top_fov, near):
        self.left_fov = left_fov
        self.right_fov = right_fov
        self.bottom_fov = bottom_fov
        self.top_fov = top_fov
        self.near = near

    def image_pyramid(self, tile_size, max_width, max_height, grid_origin):
        self.tile_size = tile_size
        self.max_width = max_width
        self.max_height = max_height
        self.grid_origin = grid_origin

    def etree_element(self):
        element = super().etree_element()
        if self._rotation:
            rotation = config.etree.SubElement(element, f"{self.ns}rotation")
            rotation.text = self._rotation
        if all(
            [
                self._left_fow,
                self._right_fov,
                self._bottom_fov,
                self._top_fov,
                self._near,
            ]
        ):
            view_volume = config.etree.SubElement(element, f"{self.ns}ViewVolume")
            left_fov = config.etree.SubElement(view_volume, f"{self.ns}leftFov")
            left_fov.text = self._left_fow
            right_fov = config.etree.SubElement(view_volume, f"{self.ns}rightFov")
            right_fov.text = self._right_fov
            bottom_fov = config.etree.SubElement(view_volume, f"{self.ns}bottomFov")
            bottom_fov.text = self._bottom_fov
            top_fov = config.etree.SubElement(view_volume, f"{self.ns}topFov")
            top_fov.text = self._top_fov
            near = config.etree.SubElement(view_volume, f"{self.ns}near")
            near.text = self._near
        if all([self._tile_size, self._max_width, self._max_height, self._grid_origin]):
            image_pyramid = config.etree.SubElement(element, f"{self.ns}ImagePyramid")
            tile_size = config.etree.SubElement(image_pyramid, f"{self.ns}tileSize")
            tile_size.text = self._tile_size
            max_width = config.etree.SubElement(image_pyramid, f"{self.ns}maxWidth")
            max_width.text = self._max_width
            max_height = config.etree.SubElement(image_pyramid, f"{self.ns}maxHeight")
            max_height.text = self._max_height
            grid_origin = config.etree.SubElement(image_pyramid, f"{self.ns}gridOrigin")
            grid_origin.text = self._grid_origin
        return element

    def from_element(self, element):
        super().from_element(element)
        rotation = element.find(f"{self.ns}rotation")
        if rotation is not None:
            self.rotation = rotation.text
        view_volume = element.find(f"{self.ns}ViewVolume")
        if view_volume is not None:
            left_fov = view_volume.find(f"{self.ns}leftFov")
            if left_fov is not None:
                self.left_fov = left_fov.text
            right_fov = view_volume.find(f"{self.ns}rightFov")
            if right_fov is not None:
                self.right_fov = right_fov.text
            bottom_fov = view_volume.find(f"{self.ns}bottomFov")
            if bottom_fov is not None:
                self.bottom_fov = bottom_fov.text
            top_fov = view_volume.find(f"{self.ns}topFov")
            if top_fov is not None:
                self.top_fov = top_fov.text
            near = view_volume.find(f"{self.ns}near")
            if near is not None:
                self.near = near.text
        image_pyramid = element.find(f"{self.ns}ImagePyramid")
        if image_pyramid is not None:
            tile_size = image_pyramid.find(f"{self.ns}tileSize")
            if tile_size is not None:
                self.tile_size = tile_size.text
            max_width = image_pyramid.find(f"{self.ns}maxWidth")
            if max_width is not None:
                self.max_width = max_width.text
            max_height = image_pyramid.find(f"{self.ns}maxHeight")
            if max_height is not None:
                self.max_height = max_height.text
            grid_origin = image_pyramid.find(f"{self.ns}gridOrigin")
            if grid_origin is not None:
                self.grid_origin = grid_origin.text
        point = element.find(f"{self.ns}Point")
        if point is not None:
            self.point = point.text
        shape = element.find(f"{self.ns}shape")
        if shape is not None:
            self.shape = shape.text


class GroundOverlay(_Overlay):
    """
    This element draws an image overlay draped onto the terrain. The <href>
    child of <Icon> specifies the image to be used as the overlay. This file
    can be either on a local file system or on a web server. If this element
    is omitted or contains no <href>, a rectangle is drawn using the color and
    LatLonBox bounds defined by the ground overlay.
    """

    __name__ = "GroundOverlay"

    _altitude = None
    # Specifies the distance above the earth's surface, in meters, and is
    # interpreted according to the altitude mode.

    _altitude_mode = "clampToGround"
    # Specifies how the <altitude> is interpreted. Possible values are:
    #   clampToGround -
    #       (default) Indicates to ignore the altitude specification and drape
    #       the overlay over the terrain.
    #   absolute -
    #       Sets the altitude of the overlay relative to sea level, regardless
    #       of the actual elevation of the terrain beneath the element. For
    #       example, if you set the altitude of an overlay to 10 meters with an
    #       absolute altitude mode, the overlay will appear to be at ground
    #       level if the terrain beneath is also 10 meters above sea level. If
    #       the terrain is 3 meters above sea level, the overlay will appear
    #       elevated above the terrain by 7 meters.

    # - LatLonBox -
    # TODO: Convert this to it's own class?
    # Specifies where the top, bottom, right, and left sides of a bounding box
    # for the ground overlay are aligned. Also, optionally the rotation of the
    # overlay.

    _north = None
    # Specifies the latitude of the north edge of the bounding box, in decimal
    # degrees from 0 to ±90.

    _south = None
    # Specifies the latitude of the south edge of the bounding box, in decimal
    # degrees from 0 to ±90.

    _east = None
    # Specifies the longitude of the east edge of the bounding box, in decimal
    # degrees from 0 to ±180. (For overlays that overlap the meridian of 180°
    # longitude, values can extend beyond that range.)

    _west = None
    # Specifies the longitude of the west edge of the bounding box, in decimal
    # degrees from 0 to ±180. (For overlays that overlap the meridian of 180°
    # longitude, values can extend beyond that range.)

    _rotation = None
    # Specifies a rotation of the overlay about its center, in degrees. Values
    # can be ±180. The default is 0 (north). Rotations are specified in a
    # counterclockwise direction.

    # TODO: <gx:LatLonQuad>
    # Used for nonrectangular quadrilateral ground overlays.
    _lat_lon_quad = None

    @property
    def altitude(self):
        return self._altitude

    @altitude.setter
    def altitude(self, value):
        if isinstance(value, (str, int, float)):
            self._altitude = str(value)
        elif value is None:
            self._altitude = None
        else:
            raise ValueError

    @property
    def altitude_mode(self):
        return self._altitude_mode

    @altitude_mode.setter
    def altitude_mode(self, mode):
        if mode in ("clampToGround", "absolute"):
            self._altitude_mode = str(mode)
        else:
            self._altitude_mode = "clampToGround"

    @property
    def north(self):
        return self._north

    @north.setter
    def north(self, value):
        if isinstance(value, (str, int, float)):
            self._north = str(value)
        elif value is None:
            self._north = None
        else:
            raise ValueError

    @property
    def south(self):
        return self._south

    @south.setter
    def south(self, value):
        if isinstance(value, (str, int, float)):
            self._south = str(value)
        elif value is None:
            self._south = None
        else:
            raise ValueError

    @property
    def east(self):
        return self._east

    @east.setter
    def east(self, value):
        if isinstance(value, (str, int, float)):
            self._east = str(value)
        elif value is None:
            self._east = None
        else:
            raise ValueError

    @property
    def west(self):
        return self._west

    @west.setter
    def west(self, value):
        if isinstance(value, (str, int, float)):
            self._west = str(value)
        elif value is None:
            self._west = None
        else:
            raise ValueError

    @property
    def rotation(self):
        return self._rotation

    @rotation.setter
    def rotation(self, value):
        if isinstance(value, (str, int, float)):
            self._rotation = str(value)
        elif value is None:
            self._rotation = None
        else:
            raise ValueError

    def lat_lon_box(
        self, north: int, south: int, east: int, west: int, rotation: int = 0
    ) -> None:
        if -90 <= float(north) <= 90:
            self.north = north
        else:
            raise ValueError
        if -90 <= float(south) <= 90:
            self.south = south
        else:
            raise ValueError
        if -180 <= float(east) <= 180:
            self.east = east
        else:
            raise ValueError
        if -180 <= float(west) <= 180:
            self.west = west
        else:
            raise ValueError
        if -180 <= float(rotation) <= 180:
            self.rotation = rotation
        else:
            raise ValueError

    def etree_element(self) -> Element:
        element = super().etree_element()
        if self._altitude:
            altitude = config.etree.SubElement(element, f"{self.ns}altitude")
            altitude.text = self._altitude
            if self._altitude_mode:
                altitude_mode = config.etree.SubElement(
                    element, f"{self.ns}altitudeMode"
                )
                altitude_mode.text = self._altitude_mode
        if all([self._north, self._south, self._east, self._west]):
            lat_lon_box = config.etree.SubElement(element, f"{self.ns}LatLonBox")
            north = config.etree.SubElement(lat_lon_box, f"{self.ns}north")
            north.text = self._north
            south = config.etree.SubElement(lat_lon_box, f"{self.ns}south")
            south.text = self._south
            east = config.etree.SubElement(lat_lon_box, f"{self.ns}east")
            east.text = self._east
            west = config.etree.SubElement(lat_lon_box, f"{self.ns}west")
            west.text = self._west
            if self._rotation:
                rotation = config.etree.SubElement(lat_lon_box, f"{self.ns}rotation")
                rotation.text = self._rotation
        return element

    def from_element(self, element: Element) -> None:
        super().from_element(element)
        altitude = element.find(f"{self.ns}altitude")
        if altitude is not None:
            self.altitude = altitude.text
        altitude_mode = element.find(f"{self.ns}altitudeMode")
        if altitude_mode is not None:
            self.altitude_mode = altitude_mode.text
        lat_lon_box = element.find(f"{self.ns}LatLonBox")
        if lat_lon_box is not None:
            north = lat_lon_box.find(f"{self.ns}north")
            if north is not None:
                self.north = north.text
            south = lat_lon_box.find(f"{self.ns}south")
            if south is not None:
                self.south = south.text
            east = lat_lon_box.find(f"{self.ns}east")
            if east is not None:
                self.east = east.text
            west = lat_lon_box.find(f"{self.ns}west")
            if west is not None:
                self.west = west.text
            rotation = lat_lon_box.find(f"{self.ns}rotation")
            if rotation is not None:
                self.rotation = rotation.text


class Document(_Container):
    """
    A Document is a container for features and styles. This element is
    required if your KML file uses shared styles or schemata for typed
    extended data
    """

    __name__ = "Document"
    _schemata = None

    def schemata(self) -> None:
        if self._schemata:
            yield from self._schemata

    def append_schema(self, schema: "Schema") -> None:
        if self._schemata is None:
            self._schemata = []
        if isinstance(schema, Schema):
            self._schemata.append(schema)
        else:
            s = Schema(schema)
            self._schemata.append(s)

    def from_element(self, element: Element) -> None:
        super().from_element(element)
        documents = element.findall(f"{self.ns}Document")
        for document in documents:
            feature = Document(self.ns)
            feature.from_element(document)
            self.append(feature)
        folders = element.findall(f"{self.ns}Folder")
        for folder in folders:
            feature = Folder(self.ns)
            feature.from_element(folder)
            self.append(feature)
        placemarks = element.findall(f"{self.ns}Placemark")
        for placemark in placemarks:
            feature = Placemark(self.ns)
            feature.from_element(placemark)
            self.append(feature)
        schemata = element.findall(f"{self.ns}Schema")
        for schema in schemata:
            s = Schema(self.ns, id="default")
            s.from_element(schema)
            self.append_schema(s)

    def etree_element(self) -> Element:
        element = super().etree_element()
        if self._schemata is not None:
            for schema in self._schemata:
                element.append(schema.etree_element())
        return element

    def get_style_by_url(self, style_url: str) -> Union[Style, StyleMap]:
        id = urlparse.urlparse(style_url).fragment
        for style in self.styles():
            if style.id == id:
                return style


class Folder(_Container):
    """
    A Folder is used to arrange other Features hierarchically
    (Folders, Placemarks, #NetworkLinks, or #Overlays).
    """

    __name__ = "Folder"

    def from_element(self, element: Element) -> None:
        super().from_element(element)
        folders = element.findall(f"{self.ns}Folder")
        for folder in folders:
            feature = Folder(self.ns)
            feature.from_element(folder)
            self.append(feature)
        placemarks = element.findall(f"{self.ns}Placemark")
        for placemark in placemarks:
            feature = Placemark(self.ns)
            feature.from_element(placemark)
            self.append(feature)
        documents = element.findall(f"{self.ns}Document")
        for document in documents:
            feature = Document(self.ns)
            feature.from_element(document)
            self.append(feature)


class Placemark(_Feature):
    """
    A Placemark is a Feature with associated Geometry.
    In Google Earth, a Placemark appears as a list item in the Places
    panel. A Placemark with a Point has an icon associated with it that
    marks a point on the Earth in the 3D viewer.
    """

    __name__ = "Placemark"
    _geometry = None

    @property
    def geometry(self):
        return self._geometry.geometry

    @geometry.setter
    def geometry(self, geometry):
        if isinstance(geometry, Geometry):
            self._geometry = geometry
        else:
            self._geometry = Geometry(ns=self.ns, geometry=geometry)

    def from_element(self, element: Element) -> None:
        super().from_element(element)
        point = element.find(f"{self.ns}Point")
        if point is not None:
            geom = Geometry(ns=self.ns)
            geom.from_element(point)
            self._geometry = geom
            return
        line = element.find(f"{self.ns}LineString")
        if line is not None:
            geom = Geometry(ns=self.ns)
            geom.from_element(line)
            self._geometry = geom
            return
        polygon = element.find(f"{self.ns}Polygon")
        if polygon is not None:
            geom = Geometry(ns=self.ns)
            geom.from_element(polygon)
            self._geometry = geom
            return
        linearring = element.find(f"{self.ns}LinearRing")
        if linearring is not None:
            geom = Geometry(ns=self.ns)
            geom.from_element(linearring)
            self._geometry = geom
            return
        multigeometry = element.find(f"{self.ns}MultiGeometry")
        if multigeometry is not None:
            geom = Geometry(ns=self.ns)
            geom.from_element(multigeometry)
            self._geometry = geom
            return
        track = element.find(f"{self.ns}Track")
        if track is not None:
            geom = gx.GxGeometry(ns=gx.NS)
            geom.from_element(track)
            self._geometry = geom
            return
        multitrack = element.find(f"{self.ns}MultiTrack")
        if line is not None:
            geom = gx.GxGeometry(ns=gx.NS)
            geom.from_element(multitrack)
            self._geometry = geom
            return
        logger.warning("No geometries found")
        logger.debug("Problem with element: %", config.etree.tostring(element))
        # raise ValueError('No geometries found')

    def etree_element(self) -> Element:
        element = super().etree_element()
        if self._geometry is not None:
            element.append(self._geometry.etree_element())
        else:
            logger.error("Object does not have a geometry")
        return element


class KML:
    """represents a KML File"""

    _features = []
    ns = None

    def __init__(self, ns: Optional[str] = None) -> None:
        """The namespace (ns) may be empty ('') if the 'kml:' prefix is
        undesired. Note that all child elements like Document or Placemark need
        to be initialized with empty namespace as well in this case.

        """
        self._features = []

        self.ns = config.KMLNS if ns is None else ns

    def from_string(self, xml_string: str) -> None:
        """create a KML object from a xml string"""
        try:
            element = config.etree.fromstring(
                xml_string, parser=config.etree.XMLParser(huge_tree=True, recover=True)
            )
        except TypeError:
            element = config.etree.XML(xml_string)

        if not element.tag.endswith("kml"):
            raise TypeError

        ns = element.tag.rstrip("kml")
        documents = element.findall(f"{ns}Document")
        for document in documents:
            feature = Document(ns)
            feature.from_element(document)
            self.append(feature)
        folders = element.findall(f"{ns}Folder")
        for folder in folders:
            feature = Folder(ns)
            feature.from_element(folder)
            self.append(feature)
        placemarks = element.findall(f"{ns}Placemark")
        for placemark in placemarks:
            feature = Placemark(ns)
            feature.from_element(placemark)
            self.append(feature)
        groundoverlays = element.findall(f"{ns}GroundOverlay")
        for groundoverlay in groundoverlays:
            feature = GroundOverlay(ns)
            feature.from_element(groundoverlay)
            self.append(feature)
        photo_overlays = element.findall(f"{ns}PhotoOverlay")
        for photo_overlay in photo_overlays:
            feature = PhotoOverlay(ns)
            feature.from_element(photo_overlay)
            self.append(feature)

    def etree_element(self) -> Element:
        # self.ns may be empty, which leads to unprefixed kml elements.
        # However, in this case the xlmns should still be mentioned on the kml
        # element, just without prefix.
        if not self.ns:
            root = config.etree.Element(f"{self.ns}kml")
            root.set("xmlns", config.KMLNS[1:-1])
        else:
            try:
                root = config.etree.Element(
                    f"{self.ns}kml", nsmap={None: self.ns[1:-1]}
                )
            except TypeError:
                root = config.etree.Element(f"{self.ns}kml")
        for feature in self.features():
            root.append(feature.etree_element())
        return root

    def to_string(self, prettyprint: bool = False) -> str:
        """Return the KML Object as serialized xml"""
        try:
            return config.etree.tostring(
                self.etree_element(),
                encoding="UTF-8",
                pretty_print=prettyprint,
            ).decode("UTF-8")
        except TypeError:
            return config.etree.tostring(self.etree_element(), encoding="UTF-8").decode(
                "UTF-8"
            )

    def features(self) -> Iterator[Union[Folder, Document, Placemark]]:
        """iterate over features"""
        for feature in self._features:
            if isinstance(feature, (Document, Folder, Placemark, _Overlay)):

                yield feature
            else:
                raise TypeError(
                    "Features must be instances of "
                    "(Document, Folder, Placemark, Overlay)"
                )

    def append(self, kmlobj: Union[Folder, Document, Placemark]) -> None:
        """append a feature"""

        if isinstance(kmlobj, (Document, Folder, Placemark, _Overlay)):
            self._features.append(kmlobj)
        else:
            raise TypeError(
                "Features must be instances of (Document, Folder, Placemark, Overlay)"
            )


__all__ = [
    "Document",
    "Folder",
    "Icon",
    "PhotoOverlay",
    "GroundOverlay",
    "KML",
    "Placemark",
]



# Copyright (C) 2012 - 2022  Christian Ledermann
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA

"""
Once you've created features within Google Earth and examined the KML
code Google Earth generates, you'll notice how styles are an important
part of how your data is displayed.
"""

import logging
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Optional
from typing import Type
from typing import Union

from typing_extensions import TypedDict

from fastkml import config
from fastkml.base import _BaseObject
from fastkml.types import Element

logger = logging.getLogger(__name__)


class StyleUrl(_BaseObject):
    """
    URL of a <Style> or <StyleMap> defined in a Document. If the style
    is in the same file, use a # reference. If the style is defined in
    an external file, use a full URL along with # referencing.
    """

    __name__ = "styleUrl"
    url = None

    def __init__(
        self,
        ns: Optional[str] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        url: Optional[str] = None,
    ) -> None:
        super().__init__(ns=ns, id=id, target_id=target_id)
        self.url = url

    def etree_element(self) -> Element:
        element = super().etree_element()
        if self.url:
            element.text = self.url
        else:
            logger.warning("StyleUrl is missing required url.")
        return element

    def from_element(self, element: Element) -> None:
        super().from_element(element)
        self.url = element.text


class _StyleSelector(_BaseObject):
    """
    This is an abstract element and cannot be used directly in a KML file.
    It is the base type for the <Style> and <StyleMap> elements. The
    StyleMap element selects a style based on the current mode of the
    Placemark. An element derived from StyleSelector is uniquely identified
    by its id and its url.
    """


class _ColorStyle(_BaseObject):
    """
    abstract element; do not create.
    This is an abstract element and cannot be used directly in a KML file.
    It provides elements for specifying the color and color mode of
    extended style types.
    subclasses are: IconStyle, LabelStyle, LineStyle, PolyStyle
    """

    id = None
    color = None
    # Color and opacity (alpha) values are expressed in hexadecimal notation.
    # The range of values for any one color is 0 to 255 (00 to ff).
    # For alpha, 00 is fully transparent and ff is fully opaque.
    # The order of expression is aabbggrr, where aa=alpha (00 to ff);
    # bb=blue (00 to ff); gg=green (00 to ff); rr=red (00 to ff).

    color_mode = None
    # Values for <colorMode> are normal (no effect) and random.
    # A value of random applies a random linear scale to the base <color>

    def __init__(
        self,
        ns: Optional[str] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        color: Optional[str] = None,
        color_mode: Optional[str] = None,
    ) -> None:
        super().__init__(ns=ns, id=id, target_id=target_id)
        self.color = color
        self.color_mode = color_mode

    def etree_element(self) -> Element:
        element = super().etree_element()
        if self.color:
            color = config.etree.SubElement(  # type: ignore[attr-defined]
                element,
                f"{self.ns}color",
            )
            color.text = self.color
        if self.color_mode:
            color_mode = config.etree.SubElement(  # type: ignore[attr-defined]
                element,
                f"{self.ns}colorMode",
            )
            color_mode.text = self.color_mode
        return element

    def from_element(self, element: Element) -> None:

        super().from_element(element)
        color_mode = element.find(f"{self.ns}colorMode")
        if color_mode is not None:
            self.color_mode = color_mode.text
        color = element.find(f"{self.ns}color")
        if color is not None:
            self.color = color.text


class HotSpot(TypedDict):
    x: float
    y: float
    xunits: str  # pixels, fraction, insetPixels
    yunits: str


class IconStyle(_ColorStyle):
    """Specifies how icons for point Placemarks are drawn"""

    __name__ = "IconStyle"
    scale = 1.0
    # Resizes the icon. (float)
    heading = None
    # Direction (that is, North, South, East, West), in degrees.
    # Default=0 (North).
    icon_href = None
    # An HTTP address or a local file specification used to load an icon.
    hot_spot = None

    def __init__(
        self,
        ns: Optional[str] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        color: Optional[str] = None,
        color_mode: Optional[str] = None,
        scale: float = 1.0,
        heading: Optional[float] = None,
        icon_href: Optional[str] = None,
        hot_spot: Optional[HotSpot] = None,
    ) -> None:
        super().__init__(
            ns=ns, id=id, target_id=target_id, color=color, color_mode=color_mode
        )

        self.scale = scale
        self.heading = heading
        self.icon_href = icon_href
        self.hot_spot = hot_spot

    def etree_element(self) -> Element:
        element = super().etree_element()
        if self.scale is not None:
            scale = config.etree.SubElement(  # type: ignore[attr-defined]
                element,
                f"{self.ns}scale",
            )
            scale.text = str(self.scale)
        if self.heading is not None:
            heading = config.etree.SubElement(  # type: ignore[attr-defined]
                element,
                f"{self.ns}heading",
            )
            heading.text = str(self.heading)
        if self.icon_href:
            icon = config.etree.SubElement(  # type: ignore[attr-defined]
                element,
                f"{self.ns}Icon",
            )
            href = config.etree.SubElement(  # type: ignore[attr-defined]
                icon,
                f"{self.ns}href",
            )
            href.text = self.icon_href
        if self.hot_spot:
            hot_spot = config.etree.SubElement(  # type: ignore[attr-defined]
                element, f"{self.ns}hotSpot"
            )
            hot_spot.attrib["x"] = str(self.hot_spot["x"])
            hot_spot.attrib["y"] = str(self.hot_spot["y"])
            hot_spot.attrib["xunits"] = str(self.hot_spot["xunits"])
            hot_spot.attrib["yunits"] = str(self.hot_spot["yunits"])
        return element

    def from_element(self, element: Element) -> None:
        super().from_element(element)
        scale = element.find(f"{self.ns}scale")
        if scale is not None:
            self.scale = float(scale.text)
        heading = element.find(f"{self.ns}heading")
        if heading is not None:
            self.heading = float(heading.text)
        icon = element.find(f"{self.ns}Icon")
        if icon is not None:
            href = icon.find(f"{self.ns}href")
            if href is not None:
                self.icon_href = href.text
        hot_spot = element.find(f"{self.ns}hotSpot")
        if hot_spot is not None:
            self.hot_spot: HotSpot = {  # type: ignore[no-redef]
                "x": hot_spot.attrib["x"],  # type: ignore[attr-defined]
                "y": hot_spot.attrib["y"],  # type: ignore[attr-defined]
                "xunits": hot_spot.attrib["xunits"],  # type: ignore[attr-defined]
                "yunits": hot_spot.attrib["yunits"],  # type: ignore[attr-defined]
            }


class LineStyle(_ColorStyle):
    """
    Specifies the drawing style (color, color mode, and line width)
    for all line geometry. Line geometry includes the outlines of
    outlined polygons and the extruded "tether" of Placemark icons
    (if extrusion is enabled).
    """

    __name__ = "LineStyle"
    width = 1.0
    # Width of the line, in pixels.

    def __init__(
        self,
        ns: Optional[str] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        color: Optional[str] = None,
        color_mode: Optional[str] = None,
        width: Union[int, float] = 1,
    ) -> None:
        super().__init__(
            ns=ns, id=id, target_id=target_id, color=color, color_mode=color_mode
        )
        self.width = width

    def etree_element(self) -> Element:
        element = super().etree_element()
        if self.width is not None:
            width = config.etree.SubElement(  # type: ignore[attr-defined]
                element,
                f"{self.ns}width",
            )
            width.text = str(self.width)
        return element

    def from_element(self, element: Element) -> None:
        super().from_element(element)
        width = element.find(f"{self.ns}width")
        if width is not None:
            self.width = float(width.text)


class PolyStyle(_ColorStyle):
    """
    Specifies the drawing style for all polygons, including polygon
    extrusions (which look like the walls of buildings) and line
    extrusions (which look like solid fences).
    """

    __name__ = "PolyStyle"
    fill = 1
    # Boolean value. Specifies whether to fill the polygon.
    outline = 1
    # Boolean value. Specifies whether to outline the polygon.
    # Polygon outlines use the current LineStyle.

    def __init__(
        self,
        ns: Optional[str] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        color: Optional[str] = None,
        color_mode: Optional[str] = None,
        fill: int = 1,
        outline: int = 1,
    ) -> None:
        super().__init__(
            ns=ns, id=id, target_id=target_id, color=color, color_mode=color_mode
        )
        self.fill = fill
        self.outline = outline

    def etree_element(self) -> Element:
        element = super().etree_element()
        if self.fill is not None:
            fill = config.etree.SubElement(  # type: ignore[attr-defined]
                element,
                f"{self.ns}fill",
            )
            fill.text = str(self.fill)
        if self.outline is not None:
            outline = config.etree.SubElement(  # type: ignore[attr-defined]
                element,
                f"{self.ns}outline",
            )
            outline.text = str(self.outline)
        return element

    def from_element(self, element: Element) -> None:
        def strtobool(val: str) -> int:
            val = val.lower()
            if val == "false":
                return 0
            if val == "true":
                return 1
            return int(float(val))

        super().from_element(element)
        fill = element.find(f"{self.ns}fill")
        if fill is not None:
            self.fill = strtobool(fill.text)
        outline = element.find(f"{self.ns}outline")
        if outline is not None:
            self.outline = strtobool(outline.text)


class LabelStyle(_ColorStyle):
    """
    Specifies how the <name> of a Feature is drawn in the 3D viewer.
    """

    __name__ = "LabelStyle"
    scale = 1.0
    # Resizes the label.

    def __init__(
        self,
        ns: Optional[str] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        color: Optional[str] = None,
        color_mode: Optional[str] = None,
        scale: float = 1.0,
    ) -> None:
        super().__init__(
            ns=ns, id=id, target_id=target_id, color=color, color_mode=color_mode
        )
        self.scale = scale

    def etree_element(self) -> Element:
        element = super().etree_element()
        if self.scale is not None:
            scale = config.etree.SubElement(  # type: ignore[attr-defined]
                element,
                f"{self.ns}scale",
            )
            scale.text = str(self.scale)
        return element

    def from_element(self, element: Element) -> None:
        super().from_element(element)
        scale = element.find(f"{self.ns}scale")
        if scale is not None:
            self.scale = float(scale.text)


class BalloonStyle(_BaseObject):
    """Specifies how the description balloon for placemarks is drawn.
    The <bgColor>, if specified, is used as the background color of
    the balloon."""

    __name__ = "BalloonStyle"

    bg_color = None
    # Background color of the balloon (optional). Color and opacity (alpha)
    # values are expressed in hexadecimal notation. The range of values for
    # any one color is 0 to 255 (00 to ff). The order of expression is
    # aabbggrr, where aa=alpha (00 to ff); bb=blue (00 to ff);
    # gg=green (00 to ff); rr=red (00 to ff).
    # For alpha, 00 is fully transparent and ff is fully opaque.
    # For example, if you want to apply a blue color with 50 percent
    # opacity to an overlay, you would specify the following:
    # <bgColor>7fff0000</bgColor>, where alpha=0x7f, blue=0xff, green=0x00,
    # and red=0x00. The default is opaque white (ffffffff).
    # Note: The use of the <color> element within <BalloonStyle> has been
    # deprecated. Use <bgColor> instead.

    text_color = None
    # Foreground color for text. The default is black (ff000000).

    text = None
    # Text displayed in the balloon. If no text is specified, Google Earth
    # draws the default balloon (with the Feature <name> in boldface,
    # the Feature <description>, links for driving directions, a white
    # background, and a tail that is attached to the point coordinates of
    # the Feature, if specified).
    # You can add entities to the <text> tag using the following format to
    # refer to a child element of Feature: $[name], $[description], $[address],
    # $[id], $[Snippet]. Google Earth looks in the current Feature for the
    # corresponding string entity and substitutes that information in the
    # balloon.
    # To include To here - From here driving directions in the balloon,
    # use the $[geDirections] tag. To prevent the driving directions links
    # from appearing in a balloon, include the <text> element with some content
    # or with $[description] to substitute the basic Feature <description>.
    # For example, in the following KML excerpt, $[name] and $[description]
    # fields will be replaced by the <name> and <description> fields found
    # in the Feature elements that use this BalloonStyle:
    # <text>This is $[name], whose description is:<br/>$[description]</text>

    display_mode = None
    # If <displayMode> is default, Google Earth uses the information supplied
    # in <text> to create a balloon . If <displayMode> is hide, Google Earth
    # does not display the balloon. In Google Earth, clicking the List View
    # icon for a Placemark whose balloon's <displayMode> is hide causes
    # Google Earth to fly to the Placemark.

    def __init__(
        self,
        ns: Optional[str] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        bg_color: Optional[str] = None,
        text_color: Optional[str] = None,
        text: Optional[str] = None,
        display_mode: Optional[str] = None,
    ) -> None:
        super().__init__(ns=ns, id=id, target_id=target_id)
        self.bg_color = bg_color
        self.text_color = text_color
        self.text = text
        self.display_mode = display_mode

    def from_element(self, element: Element) -> None:
        super().from_element(element)
        bg_color = element.find(f"{self.ns}bgColor")
        if bg_color is not None:
            self.bg_color = bg_color.text
        else:
            bg_color = element.find(f"{self.ns}color")  # type: ignore[unreachable]
            if bg_color is not None:
                self.bg_color = bg_color.text
        text_color = element.find(f"{self.ns}textColor")
        if text_color is not None:
            self.text_color = text_color.text
        text = element.find(f"{self.ns}text")
        if text is not None:
            self.text = text.text
        display_mode = element.find(f"{self.ns}displayMode")
        if display_mode is not None:
            self.display_mode = display_mode.text

    def etree_element(self) -> Element:
        element = super().etree_element()
        if self.bg_color is not None:
            elem = config.etree.SubElement(  # type: ignore[attr-defined]
                element,
                f"{self.ns}bgColor",
            )
            elem.text = self.bg_color
        if self.text_color is not None:
            elem = config.etree.SubElement(  # type: ignore[attr-defined]
                element,
                f"{self.ns}textColor",
            )
            elem.text = self.text_color
        if self.text is not None:
            elem = config.etree.SubElement(  # type: ignore[attr-defined]
                element,
                f"{self.ns}text",
            )
            elem.text = self.text
        if self.display_mode is not None:
            elem = config.etree.SubElement(  # type: ignore[attr-defined]
                element,
                f"{self.ns}displayMode",
            )
            elem.text = self.display_mode
        return element


AnyStyle = Union[BalloonStyle, IconStyle, LabelStyle, LineStyle, PolyStyle]


class Style(_StyleSelector):
    """
    A Style defines an addressable style group that can be referenced
    by StyleMaps and Features. Styles affect how Geometry is presented
    in the 3D viewer and how Features appear in the Places panel of the
    List view. Shared styles are collected in a <Document> and must have
    an id defined for them so that they can be referenced by the
    individual Features that use them.
    """

    __name__ = "Style"

    def __init__(
        self,
        ns: Optional[str] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        styles: Optional[Iterable[AnyStyle]] = None,
    ) -> None:
        super().__init__(ns, id, target_id)
        self._styles: List[AnyStyle] = []
        if styles:
            for style in styles:
                self.append_style(style)

    def append_style(
        self,
        style: AnyStyle,
    ) -> None:
        if isinstance(style, (_ColorStyle, BalloonStyle)):
            self._styles.append(style)
        else:
            raise TypeError

    def styles(
        self,
    ) -> Iterator[AnyStyle]:
        for style in self._styles:
            if isinstance(style, (_ColorStyle, BalloonStyle)):
                yield style
            else:
                raise TypeError

    def _get_style(self, element: Element, style_class: Type[AnyStyle]) -> None:
        style = element.find(f"{self.ns}{style_class.__name__}")
        if style is not None:
            thestyle = style_class(self.ns)
            thestyle.from_element(style)
            self.append_style(thestyle)

    def from_element(self, element: Element) -> None:
        super().from_element(element)
        for style in [BalloonStyle, IconStyle, LabelStyle, LineStyle, PolyStyle]:
            self._get_style(element, style)

    def etree_element(self) -> Element:
        element = super().etree_element()
        for style in self.styles():
            element.append(style.etree_element())
        return element


class StyleMap(_StyleSelector):
    """
    A <StyleMap> maps between two different Styles. Typically a
    <StyleMap> element is used to provide separate normal and highlighted
    styles for a placemark, so that the highlighted version appears when
    the user mouses over the icon in Google Earth.
    """

    __name__ = "StyleMap"
    normal = None
    highlight = None

    def __init__(
        self,
        ns: Optional[str] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        normal: Optional[Union[Style, StyleUrl]] = None,
        highlight: Optional[Union[Style, StyleUrl]] = None,
    ) -> None:
        super().__init__(ns, id, target_id)
        self.normal = normal
        self.highlight = highlight

    def from_element(self, element: Element) -> None:
        super().from_element(element)
        pairs = element.findall(f"{self.ns}Pair")
        for pair in pairs:
            key = pair.find(f"{self.ns}key")
            style = pair.find(f"{self.ns}Style")
            style_url = pair.find(f"{self.ns}styleUrl")
            if key.text == "highlight":
                if style is not None:
                    highlight = Style(self.ns)
                    highlight.from_element(style)
                elif style_url is not None:  # type: ignore[unreachable]
                    highlight = StyleUrl(self.ns)
                    highlight.from_element(style_url)
                else:
                    raise ValueError
                self.highlight = highlight
            elif key.text == "normal":
                if style is not None:
                    normal = Style(self.ns)
                    normal.from_element(style)
                elif style_url is not None:  # type: ignore[unreachable]
                    normal = StyleUrl(self.ns)
                    normal.from_element(style_url)
                else:
                    raise ValueError
                self.normal = normal
            else:
                raise ValueError

    def etree_element(self) -> Element:
        element = super().etree_element()
        if self.normal and isinstance(self.normal, (Style, StyleUrl)):
            pair = config.etree.SubElement(  # type: ignore[attr-defined]
                element,
                f"{self.ns}Pair",
            )
            key = config.etree.SubElement(  # type: ignore[attr-defined]
                pair,
                f"{self.ns}key",
            )
            key.text = "normal"
            pair.append(self.normal.etree_element())
        if self.highlight and isinstance(self.highlight, (Style, StyleUrl)):
            pair = config.etree.SubElement(  # type: ignore[attr-defined]
                element,
                f"{self.ns}Pair",
            )
            key = config.etree.SubElement(  # type: ignore[attr-defined]
                pair,
                f"{self.ns}key",
            )
            key.text = "highlight"
            pair.append(self.highlight.etree_element())
        return element


__all__ = [
    "BalloonStyle",
    "IconStyle",
    "LabelStyle",
    "LineStyle",
    "PolyStyle",
    "Style",
    "StyleMap",
    "StyleUrl",
]



# Copyright (C) 2012 - 2022  Christian Ledermann
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA

"""Types for fastkml."""
from typing import Callable
from typing import Iterable
from typing import Optional

from typing_extensions import Protocol
from typing_extensions import TypedDict

__all__ = ["Element", "KmlObjectMap"]


class Element(Protocol):
    """Protocol for Element."""

    tag: str
    text: str

    def set(self, tag: str, value: str) -> None:
        """Set the value of the tag."""

    def get(self, tag: str) -> str:
        """Get the value of the tag."""

    def find(self, tag: str) -> "Element":
        """Find the first element with the given tag."""

    def findall(self, tag: str) -> Iterable["Element"]:
        """Find all elements with the given tag."""

    def append(self, element: "Element") -> None:
        """Append an element to the current element."""


class KmlObjectMap(TypedDict):
    """TypedDict for KmlObjectMap."""

    kml_attr: str
    obj_attr: str
    required: bool
    from_kml: Callable[..., None]
    to_kml: Callable[..., None]
    validator: Optional[Callable[..., bool]]



# Copyright (C) 2012  Christian Ledermann
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA

import logging
import re
from typing import Any
from typing import List
from typing import Optional
from typing import Sequence
from typing import Union
from typing import cast

from pygeoif.factories import shape
from pygeoif.geometry import GeometryCollection
from pygeoif.geometry import LinearRing
from pygeoif.geometry import LineString
from pygeoif.geometry import MultiLineString
from pygeoif.geometry import MultiPoint
from pygeoif.geometry import MultiPolygon
from pygeoif.geometry import Point
from pygeoif.geometry import Polygon
from pygeoif.types import PointType

from fastkml import config
from fastkml.base import _BaseObject
from fastkml.types import Element

logger = logging.getLogger(__name__)

GeometryType = Union[Polygon, LineString, LinearRing, Point]
MultiGeometryType = Union[MultiPoint, MultiLineString, MultiPolygon, GeometryCollection]
AnyGeometryType = Union[GeometryType, MultiGeometryType]


class Geometry(_BaseObject):
    """ """

    geometry = None
    extrude = False
    tessellate = False
    altitude_mode = None

    def __init__(
        self,
        ns: Optional[str] = None,
        id: Optional[str] = None,
        geometry: Optional[Any] = None,
        extrude: bool = False,
        tessellate: bool = False,
        altitude_mode: Optional[str] = None,
    ) -> None:
        """
        geometry: a geometry that implements the __geo_interface__ convention

        extrude: boolean --> Specifies whether to connect the feature to
            the ground with a line. To extrude a Feature, the value for
            'altitudeMode' must be either relativeToGround, relativeToSeaFloor,
            or absolute. The feature is extruded toward the center of the
            Earth's sphere.
        tessellate: boolean -->  Specifies whether to allow the LineString
            to follow the terrain. To enable tessellation, the altitude
            mode must be clampToGround or clampToSeaFloor. Very large
            LineStrings should enable tessellation so that they follow
            the curvature of the earth (otherwise, they may go underground
            and be hidden).
            This field is not used by Polygon or Point. To allow a Polygon
            to follow the terrain (that is, to enable tessellation) specify
            an altitude mode of clampToGround or clampToSeaFloor.
        altitudeMode: [clampToGround, relativeToGround, absolute] -->
            Specifies how altitude components in the <coordinates> element
            are interpreted. Possible values are
                clampToGround - (default) Indicates to ignore an altitude
                    specification.
                relativeToGround - Sets the altitude of the element relative
                    to the actual ground elevation of a particular location.
                    For example, if the ground elevation of a location is
                    exactly at sea level and the altitude for a point is
                    set to 9 meters, then the elevation for the icon of a
                    point placemark elevation is 9 meters with this mode.
                    However, if the same coordinate is set over a location
                    where the ground elevation is 10 meters above sea level,
                    then the elevation of the coordinate is 19 meters.
                    A typical use of this mode is for placing telephone
                    poles or a ski lift.
                absolute - Sets the altitude of the coordinate relative to
                    sea level, regardless of the actual elevation of the
                    terrain beneath the element. For example, if you set
                    the altitude of a coordinate to 10 meters with an
                    absolute altitude mode, the icon of a point placemark
                    will appear to be at ground level if the terrain beneath
                    is also 10 meters above sea level. If the terrain is
                    3 meters above sea level, the placemark will appear
                    elevated above the terrain by 7 meters. A typical use
                    of this mode is for aircraft placement.

        https://developers.google.com/kml/documentation/kmlreference#geometry
        """
        super().__init__(ns, id)
        self.extrude = extrude
        self.tessellate = tessellate
        self.altitude_mode = altitude_mode
        if geometry:
            if isinstance(
                geometry,
                (
                    Point,
                    LineString,
                    Polygon,
                    MultiPoint,
                    MultiLineString,
                    MultiPolygon,
                    LinearRing,
                    GeometryCollection,
                ),
            ):
                self.geometry = geometry
            else:
                self.geometry = shape(geometry)

    # write kml

    def _set_altitude_mode(self, element: Element) -> None:
        if self.altitude_mode:
            # XXX add 'relativeToSeaFloor', 'clampToSeaFloor',
            assert self.altitude_mode in [
                "clampToGround",
                "relativeToGround",
                "absolute",
            ]
            if self.altitude_mode != "clampToGround":
                am_element = config.etree.SubElement(  # type: ignore[attr-defined]
                    element, f"{self.ns}altitudeMode"
                )
                am_element.text = self.altitude_mode

    def _set_extrude(self, element: Element) -> None:
        if self.extrude and self.altitude_mode in [
            "relativeToGround",
            # 'relativeToSeaFloor',
            "absolute",
        ]:
            et_element = cast(
                Element,
                config.etree.SubElement(  # type: ignore[attr-defined]
                    element, f"{self.ns}extrude"
                ),
            )
            et_element.text = "1"

    def _etree_coordinates(
        self,
        coordinates: Sequence[PointType],
    ) -> Element:
        element = cast(
            Element,
            config.etree.Element(f"{self.ns}coordinates"),  # type: ignore[attr-defined]
        )
        if len(coordinates[0]) == 2:
            if config.FORCE3D:  # and not clampToGround:
                tuples = (f"{c[0]:f},{c[1]:f},0.000000" for c in coordinates)
            else:
                tuples = (f"{c[0]:f},{c[1]:f}" for c in coordinates)
        elif len(coordinates[0]) == 3:
            tuples = (
                f"{c[0]:f},{c[1]:f},{c[2]:f}" for c in coordinates  # type: ignore[misc]
            )
        else:
            raise ValueError("Invalid dimensions")
        element.text = " ".join(tuples)
        return element

    def _etree_point(self, point: Point) -> Element:
        element = self._extrude_and_altitude_mode("Point")
        return self._extracted_from__etree_linearring_5(point, element)

    def _etree_linestring(self, linestring: LineString) -> Element:
        element = self._extrude_and_altitude_mode("LineString")
        if self.tessellate and self.altitude_mode in [
            "clampToGround",
            "clampToSeaFloor",
        ]:
            ts_element = config.etree.SubElement(  # type: ignore[attr-defined]
                element, f"{self.ns}tessellate"
            )
            ts_element.text = "1"
        return self._extracted_from__etree_linearring_5(linestring, element)

    def _etree_linearring(self, linearring: LinearRing) -> Element:
        element = self._extrude_and_altitude_mode("LinearRing")
        return self._extracted_from__etree_linearring_5(linearring, element)

    def _extracted_from__etree_linearring_5(
        self, arg0: Union[LineString, LinearRing, Point], element: Element
    ) -> Element:
        coords = list(arg0.coords)
        element.append(self._etree_coordinates(coords))
        return element

    def _etree_polygon(self, polygon: Polygon) -> Element:
        element = self._extrude_and_altitude_mode("Polygon")
        outer_boundary = cast(
            Element,
            config.etree.SubElement(  # type: ignore[attr-defined]
                element,
                f"{self.ns}outerBoundaryIs",
            ),
        )
        outer_boundary.append(self._etree_linearring(polygon.exterior))
        for ib in polygon.interiors:
            inner_boundary = cast(
                Element,
                config.etree.SubElement(  # type: ignore[attr-defined]
                    element,
                    f"{self.ns}innerBoundaryIs",
                ),
            )
            inner_boundary.append(self._etree_linearring(ib))
        return element

    def _extrude_and_altitude_mode(self, kml_geometry: str) -> Element:
        result = cast(
            Element,
            config.etree.Element(  # type: ignore[attr-defined]
                f"{self.ns}{kml_geometry}"
            ),
        )
        self._set_extrude(result)
        self._set_altitude_mode(result)
        return result

    def _etree_multipoint(self, points: MultiPoint) -> Element:
        element = cast(
            Element,
            config.etree.Element(  # type: ignore[attr-defined]
                f"{self.ns}MultiGeometry"
            ),
        )
        for point in points.geoms:
            element.append(self._etree_point(point))
        return element

    def _etree_multilinestring(self, linestrings: MultiLineString) -> Element:
        element = cast(
            Element,
            config.etree.Element(  # type: ignore[attr-defined]
                f"{self.ns}MultiGeometry"
            ),
        )
        for linestring in linestrings.geoms:
            element.append(self._etree_linestring(linestring))
        return element

    def _etree_multipolygon(self, polygons: MultiPolygon) -> Element:
        element = cast(
            Element,
            config.etree.Element(  # type: ignore[attr-defined]
                f"{self.ns}MultiGeometry"
            ),
        )
        for polygon in polygons.geoms:
            element.append(self._etree_polygon(polygon))
        return element

    def _etree_collection(self, features: GeometryCollection) -> Element:
        element = cast(
            Element,
            config.etree.Element(  # type: ignore[attr-defined]
                f"{self.ns}MultiGeometry"
            ),
        )
        for feature in features.geoms:
            if feature.geom_type == "Point":
                element.append(self._etree_point(cast(Point, feature)))
            elif feature.geom_type == "LinearRing":
                element.append(self._etree_linearring(cast(LinearRing, feature)))
            elif feature.geom_type == "LineString":
                element.append(self._etree_linestring(cast(LineString, feature)))
            elif feature.geom_type == "Polygon":
                element.append(self._etree_polygon(cast(Polygon, feature)))
            else:
                raise ValueError("Illegal geometry type.")
        return element

    def etree_element(self) -> Element:
        if isinstance(self.geometry, Point):
            return self._etree_point(self.geometry)
        elif isinstance(self.geometry, LinearRing):
            return self._etree_linearring(self.geometry)
        elif isinstance(self.geometry, LineString):
            return self._etree_linestring(self.geometry)
        elif isinstance(self.geometry, Polygon):
            return self._etree_polygon(self.geometry)
        elif isinstance(self.geometry, MultiPoint):
            return self._etree_multipoint(self.geometry)
        elif isinstance(self.geometry, MultiLineString):
            return self._etree_multilinestring(self.geometry)
        elif isinstance(self.geometry, MultiPolygon):
            return self._etree_multipolygon(self.geometry)
        elif isinstance(self.geometry, GeometryCollection):
            return self._etree_collection(self.geometry)
        else:
            raise ValueError("Illegal geometry type.")

    # read kml

    def _get_geometry_spec(self, element: Element) -> None:
        extrude = element.find(f"{self.ns}extrude")
        if extrude is not None:
            try:
                et = bool(int(extrude.text.strip()))
            except ValueError:
                et = False
            self.extrude = et
        else:
            self.extrude = False  # type: ignore[unreachable]
        tessellate = element.find(f"{self.ns}tessellate")
        if tessellate is not None:
            try:
                te = bool(int(tessellate.text.strip()))
            except ValueError:
                te = False
            self.tessellate = te
        else:
            self.tessellate = False  # type: ignore[unreachable]
        altitude_mode = element.find(f"{self.ns}altitudeMode")
        if altitude_mode is not None:
            am = altitude_mode.text.strip()
            if am in [
                "clampToGround",
                # 'relativeToSeaFloor', 'clampToSeaFloor',
                "relativeToGround",
                "absolute",
            ]:
                self.altitude_mode = am
            else:
                self.altitude_mode = None
        else:
            self.altitude_mode = None  # type: ignore[unreachable]

    def _get_coordinates(self, element: Element) -> List[PointType]:
        coordinates = element.find(f"{self.ns}coordinates")
        if coordinates is not None:
            # https://developers.google.com/kml/documentation/kmlreference#coordinates
            # Coordinates can be any number of tuples separated by a
            # space (potentially any number of whitespace characters).
            # Values in tuples should be separated by commas with no
            # spaces. Clean up badly formatted tuples by stripping
            # space following commas.
            latlons = re.sub(r", +", ",", coordinates.text.strip()).split()
            return [
                cast(PointType, tuple(float(c) for c in latlon.split(",")))
                for latlon in latlons
            ]

    def _get_linear_ring(self, element: Element) -> Optional[LinearRing]:
        # LinearRing in polygon
        lr = element.find(f"{self.ns}LinearRing")
        if lr is not None:
            coords = self._get_coordinates(lr)
            return LinearRing(coords)
        return None  # type: ignore[unreachable]

    def _get_geometry(self, element: Element) -> Optional[GeometryType]:
        # Point, LineString,
        # Polygon, LinearRing
        if element.tag == f"{self.ns}Point":
            coords = self._get_coordinates(element)
            self._get_geometry_spec(element)
            return Point.from_coordinates(coords)
        if element.tag == f"{self.ns}LineString":
            coords = self._get_coordinates(element)
            self._get_geometry_spec(element)
            return LineString(coords)
        if element.tag == f"{self.ns}Polygon":
            self._get_geometry_spec(element)
            outer_boundary = element.find(f"{self.ns}outerBoundaryIs")
            ob = self._get_linear_ring(outer_boundary)
            if not ob:
                return None
            inner_boundaries = element.findall(f"{self.ns}innerBoundaryIs")
            ibs = [
                self._get_linear_ring(inner_boundary)
                for inner_boundary in inner_boundaries
            ]
            return Polygon.from_linear_rings(ob, *[b for b in ibs if b])
        if element.tag == f"{self.ns}LinearRing":
            coords = self._get_coordinates(element)
            self._get_geometry_spec(element)
            return LinearRing(coords)
        return None

    def _get_multigeometry(self, element: Element) -> Optional[MultiGeometryType]:
        # MultiGeometry
        geoms: List[Union[AnyGeometryType, None]] = []
        if element.tag == f"{self.ns}MultiGeometry":
            points = element.findall(f"{self.ns}Point")
            for point in points:
                self._get_geometry_spec(point)
                geoms.append(Point.from_coordinates(self._get_coordinates(point)))
            linestrings = element.findall(f"{self.ns}LineString")
            for ls in linestrings:
                self._get_geometry_spec(ls)
                geoms.append(LineString(self._get_coordinates(ls)))
            polygons = element.findall(f"{self.ns}Polygon")
            for polygon in polygons:
                self._get_geometry_spec(polygon)
                outer_boundary = polygon.find(f"{self.ns}outerBoundaryIs")
                ob = self._get_linear_ring(outer_boundary)
                if not ob:
                    continue
                inner_boundaries = polygon.findall(f"{self.ns}innerBoundaryIs")
                inner_bs = [
                    self._get_linear_ring(inner_boundary)
                    for inner_boundary in inner_boundaries
                ]
                ibs: List[LinearRing] = [ib for ib in inner_bs if ib]
                geoms.append(Polygon.from_linear_rings(ob, *ibs))
            linearings = element.findall(f"{self.ns}LinearRing")
            if linearings:
                for lr in linearings:
                    self._get_geometry_spec(lr)
                    geoms.append(LinearRing(self._get_coordinates(lr)))
        clean_geoms: List[AnyGeometryType] = [g for g in geoms if g]
        if clean_geoms:
            geom_types = {geom.geom_type for geom in clean_geoms}
            if len(geom_types) > 1:
                return GeometryCollection(
                    clean_geoms,  # type: ignore[arg-type]
                )
            if "Point" in geom_types:
                return MultiPoint.from_points(
                    *clean_geoms,  # type: ignore[arg-type]
                )
            elif "LineString" in geom_types:
                return MultiLineString.from_linestrings(
                    *clean_geoms,  # type: ignore[arg-type]
                )
            elif "Polygon" in geom_types:
                return MultiPolygon.from_polygons(
                    *clean_geoms,  # type: ignore[arg-type]
                )
            elif "LinearRing" in geom_types:
                return GeometryCollection(
                    clean_geoms,  # type: ignore[arg-type]
                )
        return None

    def from_element(self, element: Element) -> None:
        geom = self._get_geometry(element)
        if geom is not None:
            self.geometry = geom
        else:
            mgeom = self._get_multigeometry(element)
            if mgeom is not None:
                self.geometry = mgeom
            else:
                logger.warning("No geometries found")


__all__ = ["Geometry"]



# Copyright (C) 2023  Christian Ledermann
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
"""Mixins for the KML classes."""
import logging
from typing import Optional

from fastkml.times import KmlDateTime
from fastkml.times import TimeSpan
from fastkml.times import TimeStamp

logger = logging.getLogger(__name__)


class TimeMixin:

    _timespan: Optional[TimeSpan] = None
    _timestamp: Optional[TimeStamp] = None

    @property
    def time_stamp(self) -> Optional[KmlDateTime]:
        """This just returns the datetime portion of the timestamp"""
        return self._timestamp.timestamp if self._timestamp is not None else None

    @time_stamp.setter
    def time_stamp(self, timestamp: Optional[KmlDateTime]) -> None:
        if self._timestamp is None:
            self._timestamp = TimeStamp(timestamp=timestamp)
        elif timestamp is None:
            self._timestamp = None
        else:
            self._timestamp.timestamp = timestamp
        if self._timespan and self._timestamp:
            logger.warning("Setting a TimeStamp, TimeSpan deleted")
            self._timespan = None

    @property
    def begin(self) -> Optional[KmlDateTime]:
        return self._timespan.begin if self._timespan is not None else None

    @begin.setter
    def begin(self, dt: Optional[KmlDateTime]) -> None:
        if self._timespan is None:
            self._timespan = TimeSpan(begin=dt)
        elif dt is None and self._timespan.end is None:
            self._timespan = None
        else:
            self._timespan.begin = dt
        if self._timespan and self._timestamp:
            logger.warning("Setting a TimeSpan, TimeStamp deleted")
            self._timestamp = None

    @property
    def end(self) -> Optional[KmlDateTime]:
        return self._timespan.end if self._timespan is not None else None

    @end.setter
    def end(self, dt: Optional[KmlDateTime]) -> None:
        if self._timespan is None:
            self._timespan = TimeSpan(end=dt)
        elif dt is None and self._timespan.begin is None:
            self._timespan = None
        else:
            self._timespan.end = dt
        if self._timespan and self._timestamp:
            logger.warning("Setting a TimeSpan, TimeStamp deleted")
            self._timestamp = None



# Copyright (C) 2012 -2022 Christian Ledermann
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA

"""
Fastkml is a library to read, write and manipulate kml files. It aims to keep
it simple and fast (using lxml if available). Fast refers to the time you spend
to write and read KML files as well as the time you spend to get aquainted to
the library or to create KML objects. It provides a subset of KML and is aimed
at documents that can be read from multiple clients such as openlayers and
google maps rather than to give you all functionality that KML on google earth
provides.
"""

from pkg_resources import DistributionNotFound
from pkg_resources import get_distribution

from fastkml.atom import Author
from fastkml.atom import Contributor
from fastkml.atom import Link
from fastkml.data import Data
from fastkml.data import ExtendedData
from fastkml.data import Schema
from fastkml.data import SchemaData
from fastkml.gx import GxGeometry
from fastkml.kml import KML
from fastkml.kml import Document
from fastkml.kml import Folder
from fastkml.kml import Placemark
from fastkml.styles import BalloonStyle
from fastkml.styles import IconStyle
from fastkml.styles import LabelStyle
from fastkml.styles import LineStyle
from fastkml.styles import PolyStyle
from fastkml.styles import Style
from fastkml.styles import StyleMap
from fastkml.styles import StyleUrl
from fastkml.times import TimeSpan
from fastkml.times import TimeStamp
from fastkml.views import Camera
from fastkml.views import LookAt

try:
    __version__ = get_distribution("fastkml").version
except DistributionNotFound:
    __version__ = "dev"

__all__ = [
    "KML",
    "Document",
    "Folder",
    "Placemark",
    "TimeSpan",
    "TimeStamp",
    "ExtendedData",
    "Data",
    "GxGeometry",
    "Schema",
    "SchemaData",
    "StyleUrl",
    "Style",
    "StyleMap",
    "IconStyle",
    "LineStyle",
    "PolyStyle",
    "LabelStyle",
    "BalloonStyle",
    "Link",
    "Author",
    "Contributor",
    "Camera",
    "LookAt",
]



# Copyright (C) 2021  Christian Ledermann
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA

"""Test the kml classes."""
from fastkml import kml
from tests.base import Lxml
from tests.base import StdLibrary


class TestStdLibrary(StdLibrary):
    """Test with the standard library."""

    def test_icon(self) -> None:
        """Test the Icon class."""
        icon = kml.Icon(
            id="icon-01",
            href="http://maps.google.com/mapfiles/kml/paddle/red-circle.png",
            refresh_mode="onInterval",
            refresh_interval=60,
            view_refresh_mode="onStop",
            view_refresh_time=4,
            view_bound_scale=1.2,
            view_format="BBOX=[bboxWest],[bboxSouth],[bboxEast],[bboxNorth]",
            http_query="clientName=fastkml",
        )

        assert icon.id == "icon-01"
        assert icon.href == "http://maps.google.com/mapfiles/kml/paddle/red-circle.png"
        assert icon.refresh_mode == "onInterval"
        assert icon.refresh_interval == 60
        assert icon.view_refresh_mode == "onStop"
        assert icon.view_refresh_time == 4
        assert icon.view_bound_scale == 1.2
        assert icon.view_format == "BBOX=[bboxWest],[bboxSouth],[bboxEast],[bboxNorth]"
        assert icon.http_query == "clientName=fastkml"

    def test_icon_read(self) -> None:
        """Test the Icon class."""
        icon = kml.Icon()
        icon.from_string(
            """
            <kml:Icon xmlns:kml="http://www.opengis.net/kml/2.2" id="icon-01">
            <kml:href>http://maps.google.com/mapfiles/kml/paddle/red-circle.png</kml:href>
            <kml:refreshMode>onInterval</kml:refreshMode>
            <kml:refreshInterval>60</kml:refreshInterval>
            <kml:viewRefreshMode>onStop</kml:viewRefreshMode>
            <kml:viewRefreshTime>4</kml:viewRefreshTime>
            <kml:viewBoundScale>1.2</kml:viewBoundScale>
            <kml:viewFormat>BBOX=[bboxWest],[bboxSouth],[bboxEast],[bboxNorth]</kml:viewFormat>
            <kml:httpQuery>clientName=fastkml</kml:httpQuery>
            </kml:Icon>
            """.strip()
        )

        assert icon.id == "icon-01"
        assert icon.href == "http://maps.google.com/mapfiles/kml/paddle/red-circle.png"
        assert icon.refresh_mode == "onInterval"
        assert icon.refresh_interval == 60
        assert icon.view_refresh_mode == "onStop"
        assert icon.view_refresh_time == 4
        assert icon.view_bound_scale == 1.2
        assert icon.view_format == "BBOX=[bboxWest],[bboxSouth],[bboxEast],[bboxNorth]"
        assert icon.http_query == "clientName=fastkml"

        icon2 = kml.Icon()
        icon2.from_string(icon.to_string())

        assert icon2.to_string() == icon.to_string()


class TestLxml(Lxml, TestStdLibrary):
    """Test with lxml."""



# Copyright (C) 2022  Christian Ledermann
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
"""Add Custom Data"""
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union
from typing import overload

from typing_extensions import TypedDict

import fastkml.config as config
from fastkml.base import _BaseObject
from fastkml.base import _XMLObject
from fastkml.types import Element


class SimpleField(TypedDict):
    name: str
    type: str
    displayName: Optional[str]  # noqa: N815


SimpleFields = List[Dict[str, str]]
SimpleFieldsListInput = List[Union[Dict[str, str], List[Dict[str, str]]]]
SimpleFieldsTupleInput = Tuple[Union[Dict[str, str], Tuple[Dict[str, str]]]]
SimpleFieldsDictInput = Dict[str, str]
SimpleFieldsInput = Optional[
    Union[SimpleFieldsListInput, SimpleFieldsTupleInput, SimpleFieldsDictInput]
]
SimpleFieldsOutput = Tuple[SimpleField, ...]


class Schema(_BaseObject):
    """
    Specifies a custom KML schema that is used to add custom data to
    KML Features.
    The "id" attribute is required and must be unique within the KML file.
    <Schema> is always a child of <Document>.
    """

    __name__ = "Schema"

    # The declaration of the custom fields, each of which must specify both the
    # type and the name of this field. If either the type or the name is
    # omitted, the field is ignored.
    name = None

    def __init__(
        self,
        ns: Optional[str] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        name: Optional[str] = None,
        fields: SimpleFieldsInput = None,
    ) -> None:
        if id is None:
            raise ValueError("Id is required for schema")
        super().__init__(ns=ns, id=id, target_id=target_id)
        self.name = name
        self._simple_fields: SimpleFields = []
        self.simple_fields = fields  # type: ignore[assignment]

    @property
    def simple_fields(self) -> SimpleFieldsOutput:
        return tuple(
            SimpleField(
                type=simple_field.get("type", ""),
                name=simple_field.get("name", ""),
                displayName=simple_field.get("displayName") or None,
            )
            for simple_field in self._simple_fields
            if simple_field.get("type") and simple_field.get("name")
        )

    @simple_fields.setter
    @overload
    def simple_fields(self, fields: SimpleFieldsListInput) -> None:
        ...

    @simple_fields.setter
    @overload
    def simple_fields(self, fields: SimpleFieldsTupleInput) -> None:
        ...

    @simple_fields.setter
    @overload
    def simple_fields(self, fields: SimpleFieldsDictInput) -> None:
        ...

    @simple_fields.setter
    def simple_fields(self, fields: SimpleFieldsInput) -> None:
        if isinstance(fields, dict):
            self.append(**fields)
        elif isinstance(fields, (list, tuple)):
            for field in fields:
                if isinstance(field, (list, tuple)):
                    self.append(*field)
                elif isinstance(field, dict):
                    self.append(**field)
        elif fields is None:
            self._simple_fields = []
        else:
            raise ValueError("Fields must be of type list, tuple or dict")

    def append(self, type: str, name: str, display_name: Optional[str] = None) -> None:
        """
        append a field.
        The declaration of the custom field, must specify both the type
        and the name of this field.
        If either the type or the name is omitted, the field is ignored.

        The type can be one of the following:
            string
            int
            uint
            short
            ushort
            float
            double
            bool

        <displayName>
        The name, if any, to be used when the field name is displayed to
        the Google Earth user. Use the [CDATA] element to escape standard
        HTML markup.
        """
        allowed_types = [
            "string",
            "int",
            "uint",
            "short",
            "ushort",
            "float",
            "double",
            "bool",
        ]
        if type not in allowed_types:
            raise TypeError(
                f"{name} has the type {type} which is invalid. "
                "The type must be one of "
                "'string', 'int', 'uint', 'short', "
                "'ushort', 'float', 'double', 'bool'"
            )
        self._simple_fields.append(
            {"type": type, "name": name, "displayName": display_name or ""}
        )

    def from_element(self, element: Element) -> None:
        super().from_element(element)
        self.name = element.get("name")
        simple_fields = element.findall(f"{self.ns}SimpleField")
        for simple_field in simple_fields:
            sfname = simple_field.get("name")
            sftype = simple_field.get("type")
            display_name = simple_field.find(f"{self.ns}displayName")
            sfdisplay_name = display_name.text if display_name is not None else None
            self.append(sftype, sfname, sfdisplay_name)

    def etree_element(self) -> Element:
        element = super().etree_element()
        if self.name:
            element.set("name", self.name)
        for simple_field in self.simple_fields:
            sf = config.etree.SubElement(  # type: ignore[attr-defined]
                element, f"{self.ns}SimpleField"
            )
            sf.set("type", simple_field["type"])
            sf.set("name", simple_field["name"])
            if simple_field.get("displayName"):
                dn = config.etree.SubElement(  # type: ignore[attr-defined]
                    sf, f"{self.ns}displayName"
                )
                dn.text = simple_field["displayName"]
        return element


class Data(_XMLObject):
    """Represents an untyped name/value pair with optional display name."""

    __name__ = "Data"

    def __init__(
        self,
        ns: Optional[str] = None,
        name: Optional[str] = None,
        value: Optional[str] = None,
        display_name: Optional[str] = None,
    ) -> None:
        super().__init__(ns)

        self.name = name
        self.value = value
        self.display_name = display_name

    def etree_element(self) -> Element:
        element = super().etree_element()
        element.set("name", self.name or "")
        value = config.etree.SubElement(  # type: ignore[attr-defined]
            element, f"{self.ns}value"
        )
        value.text = self.value
        if self.display_name:
            display_name = config.etree.SubElement(  # type: ignore[attr-defined]
                element, f"{self.ns}displayName"
            )
            display_name.text = self.display_name
        return element

    def from_element(self, element: Element) -> None:
        super().from_element(element)
        self.name = element.get("name")
        tmp_value = element.find(f"{self.ns}value")
        if tmp_value is not None:
            self.value = tmp_value.text
        display_name = element.find(f"{self.ns}displayName")
        if display_name is not None:
            self.display_name = display_name.text


class ExtendedData(_XMLObject):
    """Represents a list of untyped name/value pairs. See docs:

    -> 'Adding Untyped Name/Value Pairs'
       https://developers.google.com/kml/documentation/extendeddata

    """

    __name__ = "ExtendedData"

    def __init__(
        self,
        ns: Optional[str] = None,
        elements: Optional[List[Union[Data, "SchemaData"]]] = None,
    ) -> None:
        super().__init__(ns)
        self.elements = elements or []

    def etree_element(self) -> Element:
        element = super().etree_element()
        for subelement in self.elements:
            element.append(subelement.etree_element())
        return element

    def from_element(self, element: Element) -> None:
        super().from_element(element)
        self.elements = []
        untyped_data = element.findall(f"{self.ns}Data")
        for ud in untyped_data:
            el_data = Data(self.ns)
            el_data.from_element(ud)
            self.elements.append(el_data)
        typed_data = element.findall(f"{self.ns}SchemaData")
        for sd in typed_data:
            el_schema_data = SchemaData(self.ns, "dummy")
            el_schema_data.from_element(sd)
            self.elements.append(el_schema_data)


SchemaDataType = List[Dict[str, Union[int, str]]]
SchemaDataListInput = List[Union[Dict[str, str], SchemaDataType]]
SchemaDataTupleInput = Tuple[Union[Dict[str, str], Tuple[Dict[str, Union[int, str]]]]]
SchemaDataDictInput = Dict[str, Union[int, str]]
SchemaDataInput = Optional[
    Union[
        SchemaDataListInput,
        SchemaDataTupleInput,
        SchemaDataDictInput,
    ]
]
SchemaDataOutput = Tuple[Dict[str, Union[int, str]], ...]


class SchemaData(_XMLObject):
    """
    <SchemaData schemaUrl="anyURI">
    This element is used in conjunction with <Schema> to add typed
    custom data to a KML Feature. The Schema element (identified by the
    schemaUrl attribute) declares the custom data type. The actual data
    objects ("instances" of the custom data) are defined using the
    SchemaData element.
    The <schemaURL> can be a full URL, a reference to a Schema ID defined
    in an external KML file, or a reference to a Schema ID defined
    in the same KML file.
    """

    __name__ = "SchemaData"

    def __init__(
        self,
        ns: Optional[str] = None,
        schema_url: Optional[str] = None,
        data: Optional[List[Dict[str, str]]] = None,
    ) -> None:
        super().__init__(ns)
        if (not isinstance(schema_url, str)) or (not schema_url):
            raise ValueError("required parameter schema_url missing")
        self.schema_url = schema_url
        self._data: SchemaDataType = []
        self.data = data  # type: ignore[assignment]

    @property
    def data(self) -> SchemaDataOutput:
        return tuple(self._data)

    @data.setter
    @overload
    def data(self, data: SchemaDataListInput) -> None:
        ...

    @data.setter
    @overload
    def data(self, data: SchemaDataTupleInput) -> None:
        ...

    @data.setter
    @overload
    def data(self, data: SchemaDataDictInput) -> None:
        ...

    @data.setter
    def data(self, data: SchemaDataInput) -> None:
        if isinstance(data, (tuple, list)):
            self._data = []
            for d in data:
                if isinstance(d, (tuple, list)):
                    self.append_data(*d)
                elif isinstance(d, dict):
                    self.append_data(**d)
        elif data is None:
            self._data = []
        else:
            raise TypeError("data must be of type tuple or list")

    def append_data(self, name: str, value: Union[int, str]) -> None:
        if isinstance(name, str) and name:
            self._data.append({"name": name, "value": value})
        else:
            raise TypeError("name must be a nonempty string")

    def etree_element(self) -> Element:
        element = super().etree_element()
        element.set("schemaUrl", self.schema_url)
        for data in self.data:
            sd = config.etree.SubElement(  # type: ignore[attr-defined]
                element, f"{self.ns}SimpleData"
            )
            sd.set("name", data["name"])
            sd.text = data["value"]
        return element

    def from_element(self, element: Element) -> None:
        super().from_element(element)
        self.data = []  # type: ignore[assignment]
        self.schema_url = element.get("schemaUrl")
        simple_data = element.findall(f"{self.ns}SimpleData")
        for sd in simple_data:
            self.append_data(sd.get("name"), sd.text)



# Copyright (C) 2022 - 2023 Christian Ledermann
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA

"""Test the times classes."""
import datetime

import pytest
from dateutil.tz import tzoffset
from dateutil.tz import tzutc

import fastkml as kml
from fastkml.enums import DateTimeResolution
from fastkml.times import KmlDateTime
from tests.base import Lxml
from tests.base import StdLibrary


class TestDateTime(StdLibrary):
    """KmlDateTime implementation is independent of XML parser."""

    def test_kml_datetime_year(self):
        dt = datetime.datetime(2000, 1, 1)

        kdt = KmlDateTime(dt, DateTimeResolution.year)

        assert kdt.resolution == DateTimeResolution.year
        assert str(kdt) == "2000"
        assert bool(kdt)

    def test_kml_datetime_year_month(self):
        dt = datetime.datetime(2000, 3, 1)

        kdt = KmlDateTime(dt, DateTimeResolution.year_month)

        assert kdt.resolution == DateTimeResolution.year_month
        assert str(kdt) == "2000-03"
        assert bool(kdt)

    def test_kml_datetime_date(self):
        dt = datetime.datetime.now()

        kdt = KmlDateTime(dt, DateTimeResolution.date)

        assert kdt.resolution == DateTimeResolution.date
        assert str(kdt) == dt.date().isoformat()
        assert bool(kdt)

    def test_kml_datetime_date_implicit(self):
        dt = datetime.date.today()

        kdt = KmlDateTime(dt)

        assert kdt.resolution == DateTimeResolution.date
        assert str(kdt) == dt.isoformat()
        assert bool(kdt)

    def test_kml_datetime_datetime(self):
        dt = datetime.datetime.now()

        kdt = KmlDateTime(dt, DateTimeResolution.datetime)

        assert kdt.resolution == DateTimeResolution.datetime
        assert str(kdt) == dt.isoformat()
        assert bool(kdt)

    def test_kml_datetime_datetime_implicit(self):
        dt = datetime.datetime.now()

        kdt = KmlDateTime(dt)

        assert kdt.resolution == DateTimeResolution.datetime
        assert str(kdt) == dt.isoformat()
        assert bool(kdt)

    def test_kml_datetime_no_datetime(self):
        """When we pass dt as None bool() should return False."""
        kdt = KmlDateTime(None)

        assert kdt.resolution == DateTimeResolution.date
        assert not bool(kdt)
        with pytest.raises(AttributeError):
            str(kdt)

    def test_parse_year(self):
        dt = KmlDateTime.parse("2000")

        assert dt.resolution == DateTimeResolution.year
        assert dt.dt == datetime.datetime(2000, 1, 1)

    def test_parse_year_0(self):
        with pytest.raises(ValueError):
            KmlDateTime.parse("0000")

    def test_parse_year_month(self):
        dt = KmlDateTime.parse("2000-03")

        assert dt.resolution == DateTimeResolution.year_month
        assert dt.dt == datetime.datetime(2000, 3, 1)

    def test_parse_year_month_no_dash(self):
        dt = KmlDateTime.parse("200004")

        assert dt.resolution == DateTimeResolution.year_month
        assert dt.dt == datetime.datetime(2000, 4, 1)

    def test_parse_year_month_0(self):
        with pytest.raises(ValueError):
            KmlDateTime.parse("2000-00")

    def test_parse_year_month_13(self):
        with pytest.raises(ValueError):
            KmlDateTime.parse("2000-13")

    def test_parse_year_month_day(self):
        dt = KmlDateTime.parse("2000-03-01")

        assert dt.resolution == DateTimeResolution.date
        assert dt.dt == datetime.datetime(2000, 3, 1)

    def test_parse_year_month_day_no_dash(self):
        dt = KmlDateTime.parse("20000401")

        assert dt.resolution == DateTimeResolution.date
        assert dt.dt == datetime.datetime(2000, 4, 1)

    def test_parse_year_month_day_0(self):
        with pytest.raises(ValueError):
            KmlDateTime.parse("2000-05-00")

    def test_parse_datetime_utc(self):
        dt = KmlDateTime.parse("1997-07-16T07:30:15Z")

        assert dt.resolution == DateTimeResolution.datetime
        assert dt.dt == datetime.datetime(1997, 7, 16, 7, 30, 15, tzinfo=tzutc())

    def test_parse_datetime_with_tz(self):
        dt = KmlDateTime.parse("1997-07-16T07:30:15+01:00")

        assert dt.resolution == DateTimeResolution.datetime
        assert dt.dt == datetime.datetime(
            1997, 7, 16, 7, 30, 15, tzinfo=tzoffset(None, 3600)
        )

    def test_parse_datetime_with_tz_no_colon(self):
        dt = KmlDateTime.parse("1997-07-16T07:30:15+0100")

        assert dt.resolution == DateTimeResolution.datetime
        assert dt.dt == datetime.datetime(
            1997, 7, 16, 7, 30, 15, tzinfo=tzoffset(None, 3600)
        )

    def test_parse_datetime_no_tz(self):
        dt = KmlDateTime.parse("1997-07-16T07:30:15")

        assert dt.resolution == DateTimeResolution.datetime
        assert dt.dt == datetime.datetime(1997, 7, 16, 7, 30, 15)

    def test_parse_datetime_empty(self):
        assert KmlDateTime.parse("") is None

    def test_parse_year_month_5(self):
        """Test that a single digit month is invalid."""
        assert KmlDateTime.parse("19973") is None


class TestStdLibrary(StdLibrary):
    """Test with the standard library."""

    def test_timestamp(self):
        now = datetime.datetime.now()
        dt = KmlDateTime(now)
        ts = kml.TimeStamp(timestamp=dt)
        assert ts.timestamp.dt == now
        assert ts.timestamp.resolution == DateTimeResolution.datetime
        assert "TimeStamp>" in str(ts.to_string())
        assert "when>" in str(ts.to_string())
        assert now.isoformat() in str(ts.to_string())
        y2k = KmlDateTime(datetime.date(2000, 1, 1))
        ts = kml.TimeStamp(timestamp=y2k)
        assert ts.timestamp == y2k
        assert "2000-01-01" in str(ts.to_string())

    def test_timespan(self):
        now = KmlDateTime(datetime.datetime.now())
        y2k = KmlDateTime(datetime.datetime(2000, 1, 1))
        ts = kml.TimeSpan(end=now, begin=y2k)
        assert ts.end == now
        assert ts.begin == y2k
        assert "TimeSpan>" in str(ts.to_string())
        assert "begin>" in str(ts.to_string())
        assert "end>" in str(ts.to_string())
        assert now.dt.isoformat() in str(ts.to_string())
        assert y2k.dt.isoformat() in str(ts.to_string())
        ts.end = None
        assert now.dt.isoformat() not in str(ts.to_string())
        assert y2k.dt.isoformat() in str(ts.to_string())
        ts.begin = None
        pytest.raises(ValueError, ts.to_string)

    def test_feature_timestamp(self):
        now = datetime.datetime.now()
        f = kml.Document()
        f.time_stamp = KmlDateTime(now)
        assert f.time_stamp.dt == now
        assert now.isoformat() in str(f.to_string())
        assert "TimeStamp>" in str(f.to_string())
        assert "when>" in str(f.to_string())
        f.time_stamp = KmlDateTime(now.date())
        assert now.date().isoformat() in str(f.to_string())
        assert now.isoformat() not in str(f.to_string())
        f.time_stamp = None
        assert "TimeStamp>" not in str(f.to_string())

    def test_feature_timespan(self):
        now = datetime.datetime.now()
        y2k = datetime.datetime(2000, 1, 1)
        f = kml.Document()
        f.begin = KmlDateTime(y2k)
        f.end = KmlDateTime(now)
        assert f.begin == KmlDateTime(y2k)
        assert f.end == KmlDateTime(now)
        assert now.isoformat() in str(f.to_string())
        assert "2000-01-01" in str(f.to_string())
        assert "TimeSpan>" in str(f.to_string())
        assert "begin>" in str(f.to_string())
        assert "end>" in str(f.to_string())
        f.end = None
        assert now.isoformat() not in str(f.to_string())
        assert "2000-01-01" in str(f.to_string())
        assert "TimeSpan>" in str(f.to_string())
        assert "begin>" in str(f.to_string())
        assert "end>" not in str(f.to_string())
        f.begin = None
        assert "TimeSpan>" not in str(f.to_string())

    def test_feature_timespan_stamp(self):
        now = datetime.datetime.now()
        y2k = datetime.date(2000, 1, 1)
        f = kml.Document()
        f.begin = KmlDateTime(y2k)
        f.end = KmlDateTime(now)
        assert now.isoformat() in str(f.to_string())
        assert "2000-01-01" in str(f.to_string())
        assert "TimeSpan>" in str(f.to_string())
        assert "begin>" in str(f.to_string())
        assert "end>" in str(f.to_string())
        assert "TimeStamp>" not in str(f.to_string())
        assert "when>" not in str(f.to_string())
        # when we set a timestamp an existing timespan will be deleted
        f.time_stamp = KmlDateTime(now)
        assert now.isoformat() in str(f.to_string())
        assert "TimeStamp>" in str(f.to_string())
        assert "when>" in str(f.to_string())
        assert "2000-01-01" not in str(f.to_string())
        assert "TimeSpan>" not in str(f.to_string())
        assert "begin>" not in str(f.to_string())
        assert "end>" not in str(f.to_string())
        # when we set a timespan an existing timestamp will be deleted
        f.end = y2k
        assert now.isoformat() not in str(f.to_string())
        assert "2000-01-01" in str(f.to_string())
        assert "TimeSpan>" in str(f.to_string())
        assert "begin>" not in str(f.to_string())
        assert "end>" in str(f.to_string())
        assert "TimeStamp>" not in str(f.to_string())
        assert "when>" not in str(f.to_string())
        # We manipulate our Feature so it has timespan and stamp
        ts = kml.TimeStamp(timestamp=now)
        f._timestamp = ts
        # this raises an exception as only either timespan or timestamp
        # are allowed not both
        pytest.raises(ValueError, f.to_string)

    def test_read_timestamp(self):
        ts = kml.TimeStamp(ns="")
        doc = """
        <TimeStamp>
          <when>1997</when>
        </TimeStamp>
        """

        ts.from_string(doc)
        assert ts.timestamp.resolution == DateTimeResolution.year
        assert ts.timestamp.dt == datetime.datetime(1997, 1, 1, 0, 0)
        doc = """
        <TimeStamp>
          <when>1997-07</when>
        </TimeStamp>
        """

        ts.from_string(doc)
        assert ts.timestamp.resolution == DateTimeResolution.year_month
        assert ts.timestamp.dt == datetime.datetime(1997, 7, 1, 0, 0)
        doc = """
        <TimeStamp>
          <when>199808</when>
        </TimeStamp>
        """

        ts.from_string(doc)
        assert ts.timestamp.resolution == DateTimeResolution.year_month
        assert ts.timestamp.dt == datetime.datetime(1998, 8, 1, 0, 0)
        doc = """
        <TimeStamp>
          <when>1997-07-16</when>
        </TimeStamp>
        """

        ts.from_string(doc)
        assert ts.timestamp.resolution == DateTimeResolution.date
        assert ts.timestamp.dt == datetime.datetime(1997, 7, 16, 0, 0)
        # dateTime (YYYY-MM-DDThh:mm:ssZ)
        # Here, T is the separator between the calendar and the hourly notation
        # of time, and Z indicates UTC. (Seconds are required.)
        doc = """
        <TimeStamp>
          <when>1997-07-16T07:30:15Z</when>
        </TimeStamp>
        """

        ts.from_string(doc)
        assert ts.timestamp.resolution == DateTimeResolution.datetime
        assert ts.timestamp.dt == datetime.datetime(
            1997, 7, 16, 7, 30, 15, tzinfo=tzutc()
        )
        doc = """
        <TimeStamp>
          <when>1997-07-16T10:30:15+03:00</when>
        </TimeStamp>
        """

        ts.from_string(doc)
        assert ts.timestamp.resolution == DateTimeResolution.datetime
        assert ts.timestamp.dt == datetime.datetime(
            1997, 7, 16, 10, 30, 15, tzinfo=tzoffset(None, 10800)
        )

    def test_read_timespan(self):
        ts = kml.TimeSpan(ns="")
        doc = """
        <TimeSpan>
            <begin>1876-08-01</begin>
            <end>1997-07-16T07:30:15Z</end>
        </TimeSpan>
        """

        ts.from_string(doc)
        assert ts.begin.resolution == DateTimeResolution.date
        assert ts.begin.dt == datetime.datetime(1876, 8, 1, 0, 0)
        assert ts.end.resolution == DateTimeResolution.datetime
        assert ts.end.dt == datetime.datetime(1997, 7, 16, 7, 30, 15, tzinfo=tzutc())

    def test_featurefromstring(self):
        d = kml.Document(ns="")
        doc = """<Document>
          <name>Document.kml</name>
          <open>1</open>
          <TimeStamp>
            <when>1997-07-16T10:30:15+03:00</when>
          </TimeStamp>
          <TimeSpan>
            <begin>1876-08-01</begin>
            <end>1997-07-16T07:30:15Z</end>
          </TimeSpan>
        </Document>"""

        d.from_string(doc)


class TestLxml(Lxml, TestStdLibrary):
    """Test with lxml."""
